# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from pathlib import Path
import MDAnalysis as mda
import numpy as np
from numpy.testing import assert_allclose, assert_equal

from MDAnalysisTests.util import no_deprecated_call
from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    ITP,  # GROMACS itp
    ITP_nomass,  # from Automated Topology Builder
    ITP_atomtypes,
    ITP_charges,
    ITP_edited,
    ITP_tip5p,
    ITP_spce,
    GMX_TOP,
    GMX_DIR,
    GMX_TOP_BAD,
    ITP_no_endif,
)


class BaseITP(ParserBase):
    parser = mda.topology.ITPParser.ITPParser
    expected_attrs = [
        "ids",
        "names",
        "types",
        "masses",
        "charges",
        "chargegroups",
        "resids",
        "resnames",
        "segids",
        "moltypes",
        "molnums",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
    ]

    guessed_attrs = [
        "elements",
    ]

    expected_n_atoms = 63
    expected_n_residues = 10
    expected_n_segments = 1

    expected_n_bonds = 0
    expected_n_angles = 0
    expected_n_dihedrals = 0
    expected_n_impropers = 0

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(filename)

    def test_bonds_total_counts(self, top):
        assert len(top.bonds.values) == self.expected_n_bonds

    def test_angles_total_counts(self, top):
        assert len(top.angles.values) == self.expected_n_angles

    def test_dihedrals_total_counts(self, top):
        assert len(top.dihedrals.values) == self.expected_n_dihedrals

    def test_impropers_total_counts(self, top):
        assert len(top.impropers.values) == self.expected_n_impropers


class TestITP(BaseITP):
    ref_filename = ITP

    expected_n_atoms = 63
    expected_n_residues = 10
    expected_n_segments = 1

    expected_n_bonds = 62
    expected_n_angles = 91
    expected_n_dihedrals = 30
    expected_n_impropers = 29

    def test_bonds_atom_counts(self, universe):
        assert len(universe.atoms[[0]].bonds) == 3
        assert len(universe.atoms[[42]].bonds) == 1

    def test_bonds_values(self, top):
        vals = top.bonds.values
        for b in ((0, 1), (0, 2), (0, 3), (3, 4)):
            assert b in vals

    def test_bonds_type(self, universe):
        assert universe.bonds[0].type == 2

    def test_angles_atom_counts(self, universe):
        assert len(universe.atoms[[0]].angles) == 5
        assert len(universe.atoms[[42]].angles) == 2

    def test_angles_values(self, top):
        vals = top.angles.values
        for b in ((1, 0, 2), (1, 0, 3), (2, 0, 3)):
            assert (b in vals) or (b[::-1] in vals)

    def test_angles_type(self, universe):
        assert universe.angles[0].type == 2

    def test_dihedrals_atom_counts(self, universe):
        assert len(universe.atoms[[0]].dihedrals) == 2

    def test_dihedrals_multiple_types(self, universe):
        ag = universe.atoms[[0, 3, 5, 7]]
        dih = universe.dihedrals.atomgroup_intersection(ag, strict=True)[0]
        assert len(dih.type) == 2

    def test_dihedrals_values(self, top):
        vals = top.dihedrals.values
        for b in ((1, 0, 3, 5), (0, 3, 5, 7)):
            assert (b in vals) or (b[::-1] in vals)

    def test_dihedrals_type(self, universe):
        assert universe.dihedrals[0].type == (1, 1)

    def test_impropers_atom_counts(self, universe):
        assert len(universe.atoms[[0]].impropers) == 1

    def test_impropers_values(self, top):
        vals = top.impropers.values
        for b in ((3, 0, 5, 4), (5, 3, 7, 6)):
            assert (b in vals) or (b[::-1] in vals)

    def test_impropers_type(self, universe):
        assert universe.impropers[0].type == 2


class TestITPNoMass(ParserBase):
    parser = mda.topology.ITPParser.ITPParser
    ref_filename = ITP_nomass
    expected_attrs = [
        "ids",
        "names",
        "types",
        "charges",
        "chargegroups",
        "resids",
        "resnames",
        "segids",
        "moltypes",
        "molnums",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
        "masses",
    ]
    guessed_attrs = [
        "elements",
    ]

    expected_n_atoms = 60
    expected_n_residues = 1
    expected_n_segments = 1

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(filename)

    def test_mass_guess(self, universe):
        assert not np.isnan(universe.atoms[0].mass)


class TestITPAtomtypes(ParserBase):
    parser = mda.topology.ITPParser.ITPParser
    ref_filename = ITP_atomtypes
    expected_attrs = [
        "ids",
        "names",
        "types",
        "charges",
        "chargegroups",
        "resids",
        "resnames",
        "masses",
        "segids",
        "moltypes",
        "molnums",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
    ]

    expected_n_atoms = 4
    expected_n_residues = 1
    expected_n_segments = 1

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(filename)

    def test_charge_parse(self, universe):
        assert_allclose(universe.atoms[0].charge, 4)
        assert_allclose(universe.atoms[1].charge, 1.1)
        assert_allclose(universe.atoms[2].charge, -3.000)
        assert_allclose(universe.atoms[3].charge, 1.0)

    def test_mass_parse_or_guess(self, universe):
        # read from [ atoms ] section
        assert_allclose(universe.atoms[0].mass, 8.0)
        # read from [ atomtypes ] section
        assert_allclose(universe.atoms[1].mass, 20.989)
        # read from [ atomtypes ] section
        assert_allclose(universe.atoms[2].mass, 20.989)
        # guessed from element
        assert_allclose(universe.atoms[3].mass, 1.008)


class TestITPCharges(ParserBase):
    parser = mda.topology.ITPParser.ITPParser
    ref_filename = ITP_charges
    expected_attrs = [
        "ids",
        "names",
        "types",
        "masses",
        "charges",
        "chargegroups",
        "resids",
        "resnames",
        "segids",
        "moltypes",
        "molnums",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
    ]
    guessed_attrs = [
        "elements",
    ]

    expected_n_atoms = 9
    expected_n_residues = 3
    expected_n_segments = 1

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(filename)

    def test_charge_parse(self, universe):
        assert_allclose(universe.atoms[0].charge, -1.0)
        assert_allclose(universe.atoms[1].charge, 0)
        assert_allclose(universe.atoms[2].charge, 0)
        assert_allclose(universe.atoms[3].charge, -1.0)

    def test_masses_are_read(self, universe):
        assert_allclose(universe.atoms.masses, [100] * 9)


class TestDifferentDirectivesITP(BaseITP):

    ref_filename = ITP_edited

    expected_n_bonds = 62 + 118
    expected_n_angles = 88
    expected_n_dihedrals = 28
    expected_n_impropers = 29

    def test_no_extra_angles(self, top):
        for a in ((57, 59, 61), (60, 59, 61), (59, 61, 62)):
            assert a not in top.angles.values

    def test_bonds_atom_counts(self, universe):
        assert len(universe.atoms[[0]].bonds) == 5
        assert len(universe.atoms[[42]].bonds) == 5

    def test_dihedrals_atom_counts(self, universe):
        assert len(universe.atoms[[0]].dihedrals) == 1

    def test_dihedrals_identity(self, universe):
        assert universe.dihedrals[0].type == (1, 1)


class TestITPNoKeywords(BaseITP):
    expected_attrs = [
        "ids",
        "names",
        "types",
        "charges",
        "chargegroups",
        "resids",
        "resnames",
        "segids",
        "moltypes",
        "molnums",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
        "masses",
    ]
    guessed_attrs = [
        "elements",
        "masses",
    ]

    """
    Test reading ITP files *without* defined keywords.

    tip5p.itp has the lines:

        #ifndef HW1_CHARGE
            #define HW1_CHARGE 0.241
        #endif

        [ atoms ]
            1       opls_118     1       SOL              OW             1       0
            2       opls_119     1       SOL             HW1             1       HW1_CHARGE
    """
    ref_filename = ITP_tip5p
    expected_n_atoms = 5
    expected_n_residues = 1
    expected_n_segments = 1

    expected_n_bonds = 2
    # FLEXIBLE not set -> SETTLE constraint -> water has no angle
    expected_n_angles = 0
    expected_n_dihedrals = 0
    expected_n_impropers = 0

    def test_whether_settles_types(self, universe):
        for param in list(universe.bonds) + list(universe.angles):
            assert param.type == "settles"

    def test_bonds_values(self, top):
        vals = top.bonds.values
        for b in [(0, 1), (0, 2)]:
            assert b in vals

    def test_defines(self, top):
        assert_allclose(top.charges.values[1], 0.241)
        assert_allclose(top.charges.values[2], 0.241)

    def test_guessed_masses(self, filename):
        u = mda.Universe(filename)
        assert_allclose(
            u.atoms.masses, [15.999, 15.999, 15.999, 15.999, 15.999]
        )


class TestITPKeywords(TestITPNoKeywords):
    """
    Test reading ITP files *with* defined keywords.
    """

    expected_n_atoms = 7
    # FLEXIBLE is set -> no SETTLE constraint -> water should have angle
    expected_n_angles = 1

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(
            filename,
            FLEXIBLE=True,
            EXTRA_ATOMS=True,
            HW1_CHARGE=1,
            HW2_CHARGE=3,
        )

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse(
                FLEXIBLE=True, EXTRA_ATOMS=True, HW1_CHARGE=1, HW2_CHARGE=3
            )

    def test_whether_settles_types(self, universe):
        for param in list(universe.bonds) + list(universe.angles):
            assert param.type == 1

    def test_angles_values(self, top):
        assert (1, 0, 2) in top.angles.values

    def test_defines(self, top):
        assert_allclose(top.charges.values[1], 1)

    def test_kwargs_overrides_defines(self, top):
        assert_allclose(top.charges.values[2], 3)


class TestNestedIfs(BaseITP):
    """
    Test reading ITP files with nested ifdef/ifndef conditions.
    """

    ref_filename = ITP_spce
    expected_n_atoms = 7
    expected_n_residues = 1
    expected_n_segments = 1

    expected_n_bonds = 2
    expected_n_angles = 0
    expected_n_dihedrals = 0
    expected_n_impropers = 0

    @pytest.fixture
    def universe(self, filename):
        return mda.Universe(
            filename, HEAVY_H=True, EXTRA_ATOMS=True, HEAVY_SIX=True
        )

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse(HEAVY_H=True, EXTRA_ATOMS=True, HEAVY_SIX=True)

    def test_heavy_atom(self, universe):
        assert universe.atoms[5].mass > 40


class TestReadTop(BaseITP):
    """
    Test reading top files
    """

    ref_filename = GMX_TOP

    expected_n_atoms = 135
    expected_n_residues = 23
    expected_n_segments = 5

    expected_n_bonds = 130
    expected_n_angles = 182
    expected_n_dihedrals = 60
    expected_n_impropers = 58

    @pytest.fixture()
    def top(self, filename):
        with self.parser(filename) as p:
            yield p.parse(include_dir=GMX_DIR)

    @pytest.fixture()
    def universe(self, filename):
        return mda.Universe(
            filename, topology_format="ITP", include_dir=GMX_DIR
        )

    def test_output(self, filename):
        """Testing the call signature"""
        with self.parser(filename) as p:
            top = p.parse(include_dir=GMX_DIR)

    def test_creates_universe(self, filename):
        """Check that Universe works with this Parser"""
        u = mda.Universe(filename, topology_format="ITP", include_dir=GMX_DIR)

    def test_guessed_attributes(self, filename):
        """check that the universe created with certain parser have the same
        guessed attributes as  when it was guessed inside the parser"""
        u = mda.Universe(filename, topology_format="ITP", include_dir=GMX_DIR)
        for attr in self.guessed_attrs:
            assert hasattr(u.atoms, attr)

    def test_sequential(self, universe):
        resids = np.array(list(range(2, 12)) + list(range(13, 23)))
        assert_equal(universe.residues.resids[:20], resids)
        assert_equal(
            universe.residues.resindices, np.arange(self.expected_n_residues)
        )
        assert_equal(universe.atoms.chargegroups[-1], 63)


class TestErrors:

    parser = mda.topology.ITPParser.ITPParser

    def test_no_include(self):
        with pytest.raises(IOError):
            with self.parser(GMX_TOP_BAD) as p:
                top = p.parse(include_dir=GMX_DIR)

    def test_missing_endif(self):
        with pytest.raises(IOError):
            with self.parser(ITP_no_endif) as p:
                top = p.parse(include_dir=GMX_DIR)


class TestRelativePath:
    def test_relstring(self, tmpdir):
        content = """ #include "../sub3/test2.itp"
        [ atoms ]
         1      H      1    SOL    HW1      1       0.41    1.00800
        """
        content2 = """[ atoms ]
         1      H      1    SOL    HW1      1       0.41    1.00800
        """
        p = tmpdir.mkdir("sub1").join("test.itp")
        p.write(content)
        p3 = tmpdir.mkdir("sub3").join("test2.itp")
        p3.write(content2)
        p2 = tmpdir.mkdir("sub2")
        p2.chdir()
        with p2.as_cwd() as pchange:
            u = mda.Universe(str("../sub1/test.itp"), format="ITP")

    def test_relpath(self, tmpdir):
        content = """
        [ atoms ]
         1      H      1    SOL    HW1      1       0.41    1.00800
        """
        p = tmpdir.mkdir("sub1").join("test.itp")
        p.write(content)
        p2 = tmpdir.mkdir("sub2")
        p2.chdir()
        with p2.as_cwd() as pchange:
            relpath = Path("../sub1/test.itp")
            u = mda.Universe(relpath, format="ITP")

    def test_relative_path(self, tmpdir):
        test_itp_content = '#include "../atoms.itp"'
        atoms_itp_content = """
        [ moleculetype ]
        UNK 3

        [ atoms ]
        1      H      1    SOL    HW1      1       0.41    1.00800
        """
        with tmpdir.as_cwd():
            with open("atoms.itp", "w") as f:
                f.write(atoms_itp_content)
            subdir = tmpdir.mkdir("subdir")
            with subdir.as_cwd():
                with open("test.itp", "w") as f:
                    f.write(test_itp_content)
                subsubdir = subdir.mkdir("subsubdir")
                with subsubdir.as_cwd():
                    u = mda.Universe("../test.itp")
                    assert len(u.atoms) == 1


def test_missing_elements_no_attribute():
    """Check that:

    1) a warning is raised if elements are missing
    2) the elements attribute is not set
    """
    wmsg = (
        "Element information is missing, elements attribute "
        "will not be populated. If needed these can be "
    )
    with pytest.warns(UserWarning, match=wmsg):
        u = mda.Universe(ITP_atomtypes)
    with pytest.raises(AttributeError):
        _ = u.atoms.elements
    with no_deprecated_call():
        mda.Universe(ITP_atomtypes)


def test_elements_deprecation_warning():
    """Test deprecation warning is present"""
    with pytest.warns(DeprecationWarning, match="removed in release 3.0"):
        mda.Universe(ITP_nomass)


def test_elements_nodeprecation_warning():
    """Test deprecation warning is not present if elements isn't guessed"""
    with pytest.warns(UserWarning) as record:
        mda.Universe(ITP_atomtypes)
    assert len(record) == 2

    warned = [warn.message.args[0] for warn in record]
    assert "Element information is missing" in warned[0]
    assert "No coordinate reader found" in warned[1]



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pickle

import numpy as np

from numpy.testing import (
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
)

from MDAnalysis.lib.formats.libmdaxdr import TRRFile, XTCFile

from MDAnalysisTests.datafiles import TRR_multi_frame, XTC_multi_frame

import pytest

XTC_OFFSETS = np.array([0, 104, 208, 312, 416, 520, 624, 728, 832, 936])
TRR_OFFSETS = np.array([0, 480, 960, 1440, 1920, 2400, 2880, 3360, 3840, 4320])


@pytest.fixture
def xtc():
    with XTCFile(XTC_multi_frame) as f:
        yield f


@pytest.fixture
def trr():
    with TRRFile(TRR_multi_frame) as f:
        yield f


@pytest.mark.parametrize(
    "fname, xdr",
    ((XTC_multi_frame, XTCFile), (TRR_multi_frame, TRRFile)),
    indirect=True,
)
class TestCommonAPI(object):
    @staticmethod
    @pytest.fixture
    def fname(request):
        return request.param

    @staticmethod
    @pytest.fixture
    def xdr(request):
        return request.param

    @staticmethod
    @pytest.fixture
    def reader(fname, xdr):
        with xdr(fname) as f:
            yield f

    def test_natoms(self, reader):
        assert reader.n_atoms == 10

    def test_over_seek(self, reader):
        with pytest.raises(EOFError):
            reader.seek(100)

    def test_read_after_iteration(self, reader):
        for frame in reader:
            pass
        with pytest.raises(EOFError):
            reader.read()

    def test_zero_based_frame_tell(self, reader):
        assert reader.tell() == 0

    def test_read_closed(self, reader):
        reader.close()
        with pytest.raises(IOError):
            reader.read()

    def test_iter(self, reader):
        for i, frame in enumerate(reader):
            assert reader.tell() == i + 1

    def test_double_iter(self, reader):
        for i, frame in enumerate(reader):
            assert reader.tell() == i + 1
        for i, frame in enumerate(reader):
            assert reader.tell() == i + 1

    def test_seek(self, reader):
        reader.seek(4)
        assert reader.tell() == 4

    def test_seek_negative(self, reader):
        with pytest.raises(IOError):
            reader.seek(-4)

    def test_len(self, reader):
        assert len(reader) == 10
        # ensure we didn't go to another frame
        assert reader.tell() == 0

    def test_raise_not_existing(self, xdr, fname):
        with pytest.raises(IOError):
            xdr("foo")

    def test_open_wrong_mode(self, xdr, fname):
        with pytest.raises(IOError):
            xdr("foo", "e")

    def test_read_write_mode_file(self, xdr, tmpdir, fname):
        fname = str(tmpdir.join("foo"))
        with xdr(fname, "w") as f:
            with pytest.raises(IOError):
                f.read()

    @staticmethod
    def _assert_compare_readers(old_reader, new_reader):
        frame = old_reader.read()
        new_frame = new_reader.read()

        assert old_reader.fname == new_reader.fname
        assert old_reader.tell() == new_reader.tell()

        assert_equal(old_reader.offsets, new_reader.offsets)
        assert_almost_equal(frame.x, new_frame.x)
        assert_almost_equal(frame.box, new_frame.box)
        assert frame.step == new_frame.step
        assert_almost_equal(frame.time, new_frame.time)

    def test_pickle(self, reader):
        mid = len(reader) // 2
        reader.seek(mid)
        new_reader = pickle.loads(pickle.dumps(reader))
        self._assert_compare_readers(reader, new_reader)

    def test_pickle_last_frame(self, reader):
        #  This is the file state when XDRReader is in its last frame.
        #  (Issue #2878)
        reader.seek(len(reader) - 1)
        _ = reader.read()
        new_reader = pickle.loads(pickle.dumps(reader))

        assert reader.fname == new_reader.fname
        assert reader.tell() == new_reader.tell()
        with pytest.raises(StopIteration):
            new_reader.read()

    def test_pickle_closed(self, reader):
        reader.seek(len(reader) - 1)
        reader.close()
        new_reader = pickle.loads(pickle.dumps(reader))

        assert reader.fname == new_reader.fname
        assert reader.tell() != new_reader.tell()

    def test_pickle_immediately(self, reader):
        new_reader = pickle.loads(pickle.dumps(reader))

        assert reader.fname == new_reader.fname
        assert reader.tell() == new_reader.tell()


@pytest.mark.parametrize(
    "xdrfile, fname, offsets",
    (
        (XTCFile, XTC_multi_frame, XTC_OFFSETS),
        (TRRFile, TRR_multi_frame, TRR_OFFSETS),
    ),
    indirect=True,
)
class TestOffsets(object):
    @staticmethod
    @pytest.fixture
    def fname(request):
        return request.param

    @staticmethod
    @pytest.fixture
    def xdrfile(request):
        return request.param

    @staticmethod
    @pytest.fixture
    def offsets(request):
        return request.param

    @staticmethod
    @pytest.fixture
    def reader(xdrfile, fname):
        with xdrfile(fname) as f:
            yield f

    def test_offset(self, reader, offsets):
        assert_array_equal(reader.offsets, offsets)

    def test_set_offsets(self, reader, offsets):
        reader.set_offsets(np.arange(len(offsets)))
        assert_array_equal(reader.offsets, np.arange(len(offsets)))

    def test_fail_wrong_offsets(self, reader, offsets):
        reader.set_offsets(np.arange(len(offsets)))
        reader.seek(5)
        with pytest.raises(IOError):
            reader.read()

    def test_seek_overflow(self, reader, offsets):
        with pytest.raises(OverflowError):
            reader._bytes_seek(2**65)

    def test_seek_tell_all(self, reader, offsets):
        for frame, offset in enumerate(offsets):
            reader.seek(frame)
            assert reader._bytes_tell() == offset

    def test_seek_tell_all_bytes(self, reader, offsets):
        for offset in offsets:
            reader._bytes_seek(offset)
            assert reader._bytes_tell() == offset

    def test_seek_tell_largefile(self, reader, offsets):
        # Seeking/telling can be done on offsets larger than the file.
        # Filesize won't change unless a write is done at the offset.
        # We attempt to go just beyond a 4GB limit.
        big_offset = 2**32 + 1000
        # It should be safe to leave the read head so far away
        # because the file is open read-only.
        reader._bytes_seek(big_offset)
        assert reader._bytes_tell() == big_offset


@pytest.mark.parametrize(
    "xdrfile, fname", ((XTCFile, XTC_multi_frame), (TRRFile, TRR_multi_frame))
)
def test_steps(xdrfile, fname):
    with xdrfile(fname) as f:
        for i, frame in enumerate(f):
            assert frame.step == i


@pytest.mark.parametrize(
    "xdrfile, fname", ((XTCFile, XTC_multi_frame), (TRRFile, TRR_multi_frame))
)
def test_time(xdrfile, fname):
    with xdrfile(fname) as f:
        for i, frame in enumerate(f):
            assert frame.time == i * 0.5


def test_box_xtc(xtc):
    box = np.eye(3) * 20
    for frame in xtc:
        assert_array_almost_equal(frame.box, box, decimal=3)


def test_xyz_xtc(xtc):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(xtc):
        assert_array_almost_equal(frame.x, ones * i, decimal=3)


def test_box_trr(trr):
    box = np.eye(3) * 20
    for frame in trr:
        assert_array_almost_equal(frame.box, box)


def test_xyz_trr(trr):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(trr):
        assert_array_almost_equal(frame.x, ones * i)


def test_velocities_trr(trr):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(trr):
        assert_array_almost_equal(frame.v, ones * i + 10)


def test_forces_trr(trr):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(trr):
        assert_array_almost_equal(frame.f, ones * i + 20)


def test_lmbda_trr(trr):
    for i, frame in enumerate(trr):
        assert_almost_equal(frame.lmbda, 0.01 * i)


@pytest.fixture
def written_xtc(tmpdir, xtc):
    fname = str(tmpdir.join("foo.xtc"))
    with XTCFile(fname, "w") as f:
        for frame in xtc:
            f.write(*frame)
    with XTCFile(fname) as f:
        yield f


def test_written_step_xtc(written_xtc):
    for i, frame in enumerate(written_xtc):
        assert frame.step == i


def test_written_time_xtc(written_xtc):
    for i, frame in enumerate(written_xtc):
        assert frame.time == i * 0.5


def test_written_prec_xtc(written_xtc):
    for frame in written_xtc:
        assert frame.prec == 1000.0


def test_written_box_xtc(written_xtc):
    box = np.eye(3) * 20
    for frame in written_xtc:
        assert_array_almost_equal(frame.box, box, decimal=3)


def test_written_xyx_xtc(written_xtc):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(written_xtc):
        assert_array_almost_equal(frame.x, ones * i, decimal=3)


@pytest.mark.parametrize(
    "dtype", (np.int32, np.int64, np.float32, np.float64, int, float)
)
def test_write_xtc_dtype(tmpdir, dtype, xtc):
    fname = str(tmpdir.join("foo.xtc"))
    with XTCFile(fname, "w") as f:
        for frame in xtc:
            x = frame.x.astype(dtype)
            box = frame.box.astype(dtype)
            f.write(
                xyz=x,
                box=box,
                step=frame.step,
                time=frame.time,
                precision=frame.prec,
            )


@pytest.mark.parametrize("array_like", (np.array, list))
def test_write_xtc_array_like(tmpdir, array_like, xtc):
    fname = str(tmpdir.join("foo.xtc"))
    with XTCFile(fname, "w") as f:
        for frame in xtc:
            x = array_like(frame.x)
            box = array_like(frame.box)
            f.write(
                xyz=x,
                box=box,
                step=frame.step,
                time=frame.time,
                precision=frame.prec,
            )


def test_write_prec(tmpdir, xtc):
    outname = str(tmpdir.join("foo.xtc"))
    with XTCFile(outname, "w") as f_out:
        assert f_out.n_atoms == 0
        frame = xtc.read()
        f_out.write(frame.x, frame.box, frame.step, frame.time, 100.0)
    xtc = XTCFile(outname)
    assert len(xtc) == 1
    frame = xtc.read()
    assert frame.prec == 100.0


def test_different_box_xtc(tmpdir, xtc):
    """test if we can write different box-sizes for different frames."""
    orig_box = None
    fname = str(tmpdir.join("foo.xtc"))
    with XTCFile(fname, "w") as f_out:
        assert f_out.n_atoms == 0
        frame = xtc.read()
        f_out.write(frame.x, frame.box, frame.step, frame.time, frame.prec)
        orig_box = frame.box.copy()
        box = frame.box.copy() + 1
        f_out.write(frame.x, box, frame.step, frame.time, frame.prec)

    with XTCFile(fname) as xtc:
        assert len(xtc) == 2
        frame_1 = xtc.read()
        frame_2 = xtc.read()
        assert_array_almost_equal(frame_1.box, orig_box)
        assert_array_almost_equal(frame_1.box + 1, frame_2.box)


def test_write_different_x_xtc(tmpdir, xtc):
    outname = str(tmpdir.join("foo.xtc"))
    with XTCFile(outname, "w") as f_out:
        assert f_out.n_atoms == 0
        frame = xtc.read()
        f_out.write(frame.x, frame.box, frame.step, frame.time, frame.prec)
        x = np.ones((xtc.n_atoms - 1, 3))
        with pytest.raises(IOError):
            f_out.write(x, frame.box, frame.step, frame.time, frame.prec)


def test_write_different_prec(tmpdir, xtc):
    outname = str(tmpdir.join("foo.xtc"))
    with XTCFile(outname, "w") as f_out:
        assert f_out.n_atoms == 0
        frame = xtc.read()
        f_out.write(frame.x, frame.box, frame.step, frame.time, frame.prec)
        with pytest.raises(IOError):
            f_out.write(frame.x, frame.box, frame.step, frame.time, 10000.0)


@pytest.fixture
def written_trr(tmpdir, trr):
    fname = str(tmpdir.join("foo.trr"))
    with TRRFile(fname, "w") as f:
        for frame in trr:
            natoms = frame.x.shape[0]
            f.write(
                frame.x,
                frame.v,
                frame.f,
                frame.box,
                frame.step,
                frame.time,
                frame.lmbda,
                natoms,
            )
    with TRRFile(fname) as f:
        yield f


def test_written_step_trr(written_trr):
    for i, frame in enumerate(written_trr):
        assert frame.step == i


def test_written_time_trr(written_trr):
    for i, frame in enumerate(written_trr):
        assert frame.time == i * 0.5


def test_written_box_trr(written_trr):
    box = np.eye(3) * 20
    for frame in written_trr:
        assert_array_almost_equal(frame.box, box)


def test_written_xyx_trr(written_trr):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(written_trr):
        assert_array_almost_equal(frame.x, ones * i)


def test_written_velocities_trr(written_trr):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(written_trr):
        assert_array_almost_equal(frame.v, ones * i + 10)


def test_written_forces_trr(written_trr):
    ones = np.ones(30).reshape(10, 3)
    for i, frame in enumerate(written_trr):
        assert_array_almost_equal(frame.f, ones * i + 20)


@pytest.mark.parametrize(
    "dtype", (np.int32, np.int64, np.float32, np.float64, int, float)
)
def test_write_trr_dtype(tmpdir, dtype, trr):
    fname = str(tmpdir.join("foo.trr"))
    with TRRFile(fname, "w") as fout:
        for frame in trr:
            natoms = frame.x.shape[0]
            x = frame.x.astype(dtype)
            v = frame.v.astype(dtype)
            f = frame.f.astype(dtype)
            box = frame.box.astype(dtype)
            fout.write(
                x, v, f, box, frame.step, frame.time, frame.lmbda, natoms
            )


@pytest.mark.parametrize("array_like", (np.array, list))
def test_write_trr_array_like(tmpdir, array_like, trr):
    fname = str(tmpdir.join("foo.trr"))
    with TRRFile(fname, "w") as fout:
        for frame in trr:
            natoms = frame.x.shape[0]
            x = array_like(frame.x)
            v = array_like(frame.v)
            f = array_like(frame.f)
            box = array_like(frame.box)
            fout.write(
                x, v, f, box, frame.step, frame.time, frame.lmbda, natoms
            )


def test_write_different_box_trr(tmpdir, trr):
    """test if we can write different box-sizes for different frames."""
    orig_box = None
    fname = str(tmpdir.join("foo.trr"))
    with TRRFile(fname, "w") as f_out:
        assert f_out.n_atoms == 0
        frame = trr.read()
        natoms = frame.x.shape[0]
        f_out.write(
            frame.x,
            frame.v,
            frame.f,
            frame.box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms,
        )
        orig_box = frame.box.copy()
        box = frame.box.copy() + 1
        f_out.write(
            frame.x,
            frame.v,
            frame.f,
            box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms,
        )

    with TRRFile(fname) as trr:
        assert len(trr) == 2
        frame_1 = trr.read()
        frame_2 = trr.read()
        assert_array_almost_equal(frame_1.box, orig_box)
        assert_array_almost_equal(frame_1.box + 1, frame_2.box)


@pytest.fixture
def trr_writer(tmpdir, trr):
    outname = str(tmpdir.join("foo.trr"))
    with TRRFile(outname, "w") as f_out:
        assert f_out.n_atoms == 0
        frame = trr.read()
        natoms = frame.x.shape[0]
        f_out.write(
            frame.x,
            frame.v,
            frame.f,
            frame.box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms,
        )
        yield frame, f_out


def test_write_different_natoms(trr_writer):
    frame, writer = trr_writer
    natoms = frame.x.shape[0]
    with pytest.raises(IOError):
        writer.write(
            frame.x,
            frame.v,
            frame.f,
            frame.box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms - 1,
        )


def test_write_different_x_trr(trr_writer):
    frame, writer = trr_writer
    natoms = frame.x.shape[0]
    x = np.ones((natoms - 1, 3))
    with pytest.raises(IOError):
        writer.write(
            x,
            frame.v,
            frame.f,
            frame.box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms,
        )


def test_write_different_v(trr_writer):
    frame, writer = trr_writer
    natoms = frame.x.shape[0]
    v = np.ones((natoms - 1, 3))
    with pytest.raises(IOError):
        writer.write(
            frame.x,
            v,
            frame.f,
            frame.box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms,
        )


def test_write_different_f(trr_writer):
    frame, writer = trr_writer
    natoms = frame.x.shape[0]
    f = np.ones((natoms - 1, 3))
    with pytest.raises(IOError):
        writer.write(
            frame.x,
            frame.v,
            f,
            frame.box,
            frame.step,
            frame.time,
            frame.lmbda,
            natoms,
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Cython wrapper for the C implementation of the Affinity Perturbation clustering algorithm.

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

"""
from ..utils import TriangularMatrix
import logging
import numpy
cimport numpy
cimport cython

numpy.import_array()

cdef extern from "ap.h":
    int CAffinityPropagation(float*, int, float, int, int, bint, long*)

@cython.boundscheck(False)
@cython.wraparound(False)
def AffinityPropagation(s, preference, float lam, int max_iterations, int convergence, int noise=1):
    """Affinity propagation clustering algorithm. This class is a Cython wrapper around the Affinity propagation algorithm, which is implement as a C library (see ap.c). The implemented algorithm is described in the paper:

    Clustering by Passing Messages Between Data Points.
    Brendan J. Frey and Delbert Dueck, University of Toronto
    Science 315, 972–976, February 2007

    Parameters
    ----------
    s : encore.utils.TriangularMatrix object
        Triangular matrix containing the similarity values for each pair of
        clustering elements. Notice that the current implementation does not
        allow for asymmetric values (i.e. similarity(a,b) is assumed to be
        equal to similarity(b,a))
    preference : numpy.array of floats or float
        Preference values, which the determine the number of clusters. If a
        single value is given, all the preference values are set to that.
        Otherwise, the list is used to set the preference values (one value per
        element, so the list must be of the same size as the number of
        elements)
    lam : float
        Floating point value that defines how much damping is applied to the
        solution at each iteration. Must be ]0,1]
    max_iterations : int
        Maximum number of iterations
    convergence : int
        Number of iterations in which the cluster centers must remain the same
        in order to reach convergence
    noise : int
        Whether to apply noise to the input s matrix, such there are no equal
        values. 1 is for yes, 0 is for no.

    Returns
    -------
    elements : list of int or None
        List of cluster-assigned elements, which can be used by
        encore.utils.ClustersCollection to generate Cluster objects. See these
        classes for more details.

    """
    cdef int cn = s.size
    cdef float cpreference = preference

    # Assign preference values to diagonal
    try:
        for i in xrange(s.size):
            s[i,i] = <float>preference[i]
    except:
        pass

    if type(preference) == float:
        for i in xrange(s.size):
            s[i,i] = <float>preference
    else:
        raise TypeError ("Preference should be of type float")

    logging.info("Preference %3.2f: starting Affinity Propagation" % (preference))

    # Prepare input and ouput arrays
    cdef numpy.ndarray[numpy.float32_t,  ndim=1] matndarray = numpy.ascontiguousarray(s._elements, dtype=numpy.float32)
    cdef numpy.ndarray[long,   ndim=1] clusters   = numpy.zeros((s.size),dtype=numpy.dtype("long"))

    # run C module Affinity Propagation
    iterations = CAffinityPropagation(<float*>matndarray.data, cn, lam, max_iterations, convergence, noise, <long*>clusters.data)

    # Provide warning in case of lack of convergence
    if iterations == 0:
        logging.info("Preference %3.2f: could not converge in %d iterations" % (preference, -iterations))
        import warnings
        warnings.warn("Clustering with preference {0:3.2f} did not fully converge in {1:d} iterations".format(preference, -iterations))

    # Find centroids
    centroids = numpy.unique(clusters)
    for k in numpy.arange(centroids.shape[0]):
        ii = numpy.where(clusters == centroids[k])[0]
        small_mat = numpy.zeros((ii.shape[0], ii.shape[0]))
        for ii1 in numpy.arange(ii.shape[0]):
            for ii2 in numpy.arange(ii.shape[0]):
                small_mat[ii1,ii2] = s[ ii[ii1], ii[ii2] ]
        j = numpy.argmax(numpy.sum(small_mat, axis=0))

        centroids[k] = ii[j]

    # Similarity to centroids
    S_centroids = numpy.zeros((s.size, centroids.shape[0]))
    for line in numpy.arange(s.size):
        for c in numpy.arange(centroids.shape[0]):
            S_centroids[line,c] = s[line, centroids[c]]

    # Center values for each observation
    c = numpy.argmax(S_centroids, axis=1)

    # Centroids should point to themselves
    c[centroids] = numpy.arange(centroids.shape[0])

    # Assign centroid indices to all observables
    clusters = centroids[c]

    logging.info("Preference %3.2f: converged in %d iterations" % (preference, iterations))

    return clusters



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
MOL2 file format --- :mod:`MDAnalysis.coordinates.MOL2`
========================================================

Classes to read Tripos_ molecule structure format (MOL2_) coordinate
and topology files. Used by the DOCK_ docking code.

.. _MOL2: http://www.tripos.com/data/support/mol2.pdf
.. _Tripos: http://www.tripos.com/
.. _DOCK: http://dock.compbio.ucsf.edu/


Classes
-------

.. autoclass:: MOL2Parser
   :members:
   :inherited-members:

"""
import os
import numpy as np

from ..lib.util import openany
from .base import TopologyReaderBase, squash_by
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Atomtypes,
    Bonds,
    Charges,
    Elements,
    Resids,
    Resnums,
    Resnames,
    Segids,
)
from ..core.topology import Topology
from ..guesser.tables import SYBYL2SYMB

import warnings


class MOL2Parser(TopologyReaderBase):
    """Reads topology from a Tripos_ MOL2_ file.

    Creates the following Attributes:
     - Atomids
     - Atomnames
     - Atomtypes
     - Charges
     - Resids
     - Resnames
     - Bonds
     - Elements


    .. note::

        By default, masses will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.


    Notes
    -----
    Elements are obtained directly from the SYBYL atom types. If some atoms have
    unknown atom types, they will be assigned an empty element record. If all
    atoms have unknown atom types, the elements attribute will not be set.

    Dealing with optional fields:

      1. ``resid`` will set to 1 when not provided.

      2. If no atoms have ``resname`` field, resnames attribute will not be set;
         If some atoms have ``resname`` while some do not,
         :exc:`ValueError` will occur.

      3. If "NO_CHARGES" shows up in "@<TRIPOS>MOLECULE" section
         and no atoms have the ``charge`` field, charges attribute will not be set;
         If "NO_CHARGES" shows up while ``charge`` field appears,
         :exc:`ValueError` will occur;
         If a charge model is specified while some atoms don't have ``charge``,
         :exc:`ValueError` will occur as well.

    Raises
    ------
    ValueError
      If some atoms have the optional field ``resname`` (aka ``subst_name``)
      while some do not.

    ValueError
      If "NO_CHARGES" shows up in "@<TRIPOS>MOLECULE" section while
      some atoms have the optional field ``charge``.

    ValueError
      If a charge model is specified in "@<TRIPOS>MOLECULE" section while
      some atoms do not have the optional field ``charge``.


    .. versionchanged:: 0.9
       Now subclasses TopologyReaderBase
    .. versionchanged:: 0.20.0
       Allows for comments at the top of the file
       Ignores status bit strings
    .. versionchanged:: 2.0.0
       Bonds attribute is not added if no bonds are present in MOL2 file
    .. versionchanged:: 2.0.0
       Parse elements from atom types.
    .. versionchanged:: 2.2.0
       Read MOL2 files with optional columns omitted.
    .. versionchanged:: 2.8.0
        Removed mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).

    """

    format = "MOL2"

    def parse(self, **kwargs):
        """Parse MOL2 file *filename* and return the dict `structure`.

        Returns
        -------
        A MDAnalysis Topology object
        """
        blocks = []

        with openany(self.filename) as f:
            for i, line in enumerate(f):
                # found new molecules
                if "@<TRIPOS>MOLECULE" in line:
                    if len(blocks):
                        break
                    blocks.append({"start_line": i, "lines": []})
                if len(blocks):
                    blocks[-1]["lines"].append(line)

        if not len(blocks):
            raise ValueError(
                "The mol2 file '{0}' needs to have at least one"
                " @<TRIPOS>MOLECULE block".format(self.filename)
            )
        block = blocks[0]

        sections = {}
        cursor = None

        for line in block["lines"]:
            if "@<TRIPOS>" in line:
                cursor = line.split("@<TRIPOS>")[1].strip().lower()
                sections[cursor] = []
                continue
            elif line.startswith("#") or line == "\n":
                continue
            sections[cursor].append(line)

        atom_lines, bond_lines = sections["atom"], sections.get("bond")

        if not len(atom_lines):
            raise ValueError(
                "The mol2 block ({0}:{1}) has no atoms".format(
                    os.path.basename(self.filename), block["start_line"]
                )
            )

        ids = []
        names = []
        types = []
        resids = []
        resnames = []
        charges = []
        has_charges = sections["molecule"][3].strip() != "NO_CHARGES"
        for a in atom_lines:
            columns = a.split()
            if len(columns) >= 9:
                aid, name, x, y, z, atom_type, resid, resname, charge = (
                    columns[:9]
                )
            elif len(columns) < 6:
                raise ValueError(
                    f"The @<TRIPOS>ATOM block in mol2 file"
                    f" {os.path.basename(self.filename)}"
                    f" should have at least 6 fields to be"
                    f" unpacked: atom_id atom_name x y z"
                    f" atom_type [subst_id[subst_name"
                    f" [charge [status_bit]]]]"
                )
            else:
                aid, name, x, y, z, atom_type = columns[:6]
                id_name_charge = [1, None, None]
                for i in range(6, len(columns)):
                    id_name_charge[i - 6] = columns[i]
                resid, resname, charge = id_name_charge
            if has_charges:
                if charge is None:
                    raise ValueError(
                        f"The mol2 file {self.filename}"
                        f" indicates a charge model"
                        f"{sections['molecule'][3]}, but"
                        f" no charge provided in line: {a}"
                    )
            else:
                if charge is not None:
                    raise ValueError(
                        f"The mol2 file {self.filename}"
                        f" indicates no charges, but charge"
                        f" {charge} provided in line: {a}."
                    )

            ids.append(aid)
            names.append(name)
            types.append(atom_type)
            resids.append(resid)
            resnames.append(resname)
            charges.append(charge)

        n_atoms = len(ids)

        validated_elements = np.empty(n_atoms, dtype="U3")
        invalid_elements = set()
        for i, at in enumerate(types):
            if at in SYBYL2SYMB:
                validated_elements[i] = SYBYL2SYMB[at]
            else:
                invalid_elements.add(at)
                validated_elements[i] = ""

        # Print single warning for all unknown elements, if any
        if invalid_elements:
            warnings.warn(
                "Unknown elements found for some "
                f"atoms: {invalid_elements}. "
                "These have been given an empty element record."
            )

        attrs = []
        attrs.append(Atomids(np.array(ids, dtype=np.int32)))
        attrs.append(Atomnames(np.array(names, dtype=object)))
        attrs.append(Atomtypes(np.array(types, dtype=object)))
        if has_charges:
            attrs.append(Charges(np.array(charges, dtype=np.float32)))

        if not np.all(validated_elements == ""):
            attrs.append(Elements(validated_elements))

        resids = np.array(resids, dtype=np.int32)
        resnames = np.array(resnames, dtype=object)

        if np.all(resnames):
            residx, resids, (resnames,) = squash_by(resids, resnames)
            n_residues = len(resids)
            attrs.append(Resids(resids))
            attrs.append(Resnums(resids.copy()))
            attrs.append(Resnames(resnames))
        elif not np.any(resnames):
            residx, resids, _ = squash_by(
                resids,
            )
            n_residues = len(resids)
            attrs.append(Resids(resids))
            attrs.append(Resnums(resids.copy()))
        else:
            raise ValueError(
                f"Some atoms in the mol2 file {self.filename}"
                f" have subst_name while some do not."
            )

        attrs.append(Segids(np.array(["SYSTEM"], dtype=object)))

        # don't add Bonds if there are none (Issue #3057)
        if bond_lines:
            bonds = []
            bondorder = []
            for b in bond_lines:
                # bond_type can be: 1, 2, am, ar
                bid, a0, a1, bond_type = b.split()[:4]

                a0, a1 = int(a0) - 1, int(a1) - 1
                bond = tuple(sorted([a0, a1]))
                bondorder.append(bond_type)
                bonds.append(bond)
            attrs.append(Bonds(bonds, order=bondorder))

        top = Topology(
            n_atoms, n_residues, 1, attrs=attrs, atom_resindex=residx
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Custom exceptions and warnings --- :mod:`MDAnalysis.exceptions`
===============================================================

"""


class SelectionError(Exception):
    """Raised when a atom selection failed."""


class NoDataError(ValueError, AttributeError):
    """Raised when empty input is not allowed or required data are missing.

    This exception is raised in the following scenarios:

    * Raised when a :class:`~MDAnalysis.core.topologyattrs.TopologyAttr`
      (e.g., bonds, charges) is not present in the data.

    * Raised when data is missing in an analysis class because the `run()`
      method has not been called, e.g.
      :class:`~MDAnalysis.analysis.polymer.PersistenceLength`.

    * Raised when timestep data is missing, such as positions, velocities,
      or forces.

    * Raised in the `nojump` transformation if there is no box information
      in the universe.

    This exception should not be raised in cases where arrays have zero width,
    or AtomGroups are empty.

    .. versionchanged:: 1.0.0
        Now a subclass of AttributeError as well as ValueError
    """


class ApplicationError(OSError):
    """Raised when an external application failed.

    The error code is specific for the application.

    .. versionadded:: 0.7.7
    """


class SelectionWarning(Warning):
    """Warning indicating a possible problem with a selection."""


class MissingDataWarning(Warning):
    """Warning indicating is that required data are missing."""


class ConversionWarning(Warning):
    """Warning indicating a problem with converting between units."""


class FileFormatWarning(Warning):
    """Warning indicating possible problems with a file format."""


class StreamWarning(Warning):
    """Warning indicating a possible problem with a stream.

    :exc:`StreamWarning` is used when streams are substituted for simple access
    by filename (see in particular
    :class:`~MDAnalysis.lib.util.NamedStream`). This does not work everywhere
    in MDAnalysis (yet).
    """


class DuplicateWarning(UserWarning):
    """Warning indicating possible problems arising from an
    :class:`~MDAnalysis.core.groups.AtomGroup` /
    :class:`~MDAnalysis.core.groups.ResidueGroup` /
    :class:`~MDAnalysis.core.groups.SegmentGroup` containing duplicate
    :class:`Atoms<MDAnalysis.core.groups.Atom>` /
    :class:`Residues<MDAnalysis.core.groups.Residue>` /
    :class:`Segments<MDAnalysis.core.groups.Segment>`.


    .. versionadded:: 0.19.0
    """



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

# TPR parser and tpr support module
# Copyright (c) 2011 Zhuyi Xue
# Released under the  GNU Public Licence, v2


"""Gromacs portable run input TPR format parser
============================================

The :mod:`~MDAnalysis.topology.TPRParser` module allows reading of a
Gromacs_ portable run input file (a `TPR file`_). Because
the file format of the TPR file is changing rapidly, not all versions
are currently supported. The known working versions and the
approximate Gromacs release numbers are listed in the table
:ref:`TPR format versions <TPR-format-table>`.

.. _`TPR-format-table`:

.. table:: TPR format versions and generations read by :func:`MDAnalysis.topology.TPRParser.parse`.

   ========== ============== ==================== =====
   TPX format TPX generation Gromacs release      read
   ========== ============== ==================== =====
   ??         ??             3.3, 3.3.1           no

   58         17             4.0, 4.0.2, 4.0.3,   yes
                             4.0.4, 4.0.5, 4.0.6,
                             4.0.7

   73         23             4.5.0, 4.5.1, 4.5.2, yes
                             4.5.3, 4.5.4, 4.5.5

   83         24             4.6, 4.6.1           yes

   100        26             5.0, 5.0.1, 5.0.2,   yes
                             5.0.3,5.0.4, 5.0.5

   103        26             5.1                  yes

   110        26             2016                 yes
   112        26             2018                 yes
   116        26             2019                 yes
   119        27             2020[*]_             yes
   122        28             2021                 yes
   127        28             2022                 yes
   129        28             2023                 yes
   133        28             2024.1               yes
   134        28             2024.4               yes
   137        28             2025.0               yes
   ========== ============== ==================== =====

.. [*] Files generated by the beta versions of Gromacs 2020 are NOT supported.
   See `Issue 2428`_ for more details.

For further discussion and notes see `Issue 2`_. Please *open a new issue* in
the `Issue Tracker`_ when a new or different TPR file format version should be
supported.

Bonded interactions available in Gromacs are described in table 5.5 of the
`Gromacs manual`_. The following ones are used to build the topology (see
`Issue 463`_):

* bonds: regular bonds (type 1), G96 bonds (type 2), Morse (type 3),
  cubic bonds (type 4), connections (type 5), harmonic potentials (type 6),
  FENE bonds (type 7), restraint potentials (type 10),
  tabulated potential with exclusion/connection (type 8),
  tabulated potential without exclusion/connection (type 9), constraints with
  exclusion/connection (type 1), constraints without exclusion/connection (type
  2), SETTLE constraints
* angles: regular angles (type 1), G96 angles (type 2), cross bond-bond
  (type3), cross-bond-angle (type 4), Urey-Bradley (type 5), quartic angles
  (type 6), restricted bending potential (type 10), tabulated angles (type 8)
* dihedrals: proper dihedrals (type 1 and type 9), Ryckaert-Bellemans dihedrals
  (type 3), Fourier dihedrals (type 5), restricted dihedrals (type 10),
  combined bending-torsion potentials (type 11), tabulated dihedral (type 8)
* impropers: improper dihedrals (type 2), periodic improper dihedrals (type 4)


Classes
-------

.. autoclass:: TPRParser
   :members:
   :inherited-members:

See Also
--------
:mod:`MDAnalysis.topology.tpr`


Development notes
-----------------

The TPR reader is a pure-python implementation of a basic TPR
parser. Currently the following sections of the topology are parsed:

* Atoms: number, name, type, resname, resid, segid, chainID, mass, charge, element
  [residue, segment, radius, bfactor, resnum, moltype]
* Bonds
* Angles
* Dihedrals
* Impropers

This tpr parser is written according to the following files

- :file:`{gromacs_dir}/src/kernel/gmxdump.c`
- :file:`{gromacs_dir}/src/gmxlib/tpxio.c` (the most important one)
- :file:`{gromacs_dir}/src/gmxlib/gmxfio_rw.c`
- :file:`{gromacs_dir}/src/gmxlib/gmxfio_xdr.c`
- :file:`{gromacs_dir}/include/gmxfiofio.h`

or their equivalent in more recent versions of Gromacs.

The function :func:`read_tpxheader` is based on the
`TPRReaderDevelopment`_ notes.  Functions with names starting with
``read_`` or ``do_`` are trying to be similar to those in
:file:`gmxdump.c` or :file:`tpxio.c`, those with ``extract_`` are new.

Versions prior to Gromacs 4.0.x are not supported.

.. Links
.. _Gromacs: http://www.gromacs.org
.. _`Gromacs manual`: http://manual.gromacs.org/documentation/5.1/manual-5.1.pdf
.. _TPR file: http://manual.gromacs.org/current/online/tpr.html
.. _`Issue Tracker`: https://github.com/MDAnalysis/mdanalysis/issues
.. _`Issue 2`: https://github.com/MDAnalysis/mdanalysis/issues/2
.. _`Issue 463`: https://github.com/MDAnalysis/mdanalysis/pull/463
.. _TPRReaderDevelopment: https://github.com/MDAnalysis/mdanalysis/wiki/TPRReaderDevelopment
.. _`Issue 2428`: https://github.com/MDAnalysis/mdanalysis/issues/2428


.. versionchanged:: 2.0.0
   The `elements` topology attribute is now exposed if at least one atom has
   a valid element symbol. In that case, atoms for which the element is not
   recognized have their element attribute set to an empty string. If none of
   the elements are recognized, then the `elements` attribute is not set in the
   topology.

.. versionchanged:: 2.7.0
   If the TPR molblock is named "Protein_chain_XXX" then we assume that XXX is 
   describing the chain of a protein (in the sense of the PDB chainID) and set
   the topology attribute `chainID` to "XXX". In all other cases, the chainID
   remains the full molblock name. The `segID` is never changed.
"""
__author__ = "Zhuyi Xue"
__copyright__ = "GNU Public Licence, v2"


from ..lib.util import openany
from .tpr import utils as tpr_utils
from .tpr import setting as S
from .base import TopologyReaderBase
from ..core.topologyattrs import Resnums

import logging

logger = logging.getLogger("MDAnalysis.topology.TPRparser")


class TPRParser(TopologyReaderBase):
    """Read topology information from a Gromacs_ `TPR file`_.

    .. _Gromacs: http://www.gromacs.org
    .. _TPR file: http://manual.gromacs.org/current/online/tpr.html
    """

    format = "TPR"

    def parse(self, tpr_resid_from_one=True, **kwargs):
        """Parse a Gromacs TPR file into a MDAnalysis internal topology structure.

        Parameters
        ----------
        tpr_resid_from_one: bool (optional)
            Toggle whether to index resids from 1 or 0 from TPR files.
            TPR files index resids from 0 by default, even though GRO and ITP
            files index from 1.

        Returns
        -------
        structure : dict


        .. versionchanged:: 1.1.0
            Added the ``tpr_resid_from_one`` keyword to control if
            resids are indexed from 0 or 1. Default ``False``.
        .. versionchanged:: 2.0.0
            Changed to ``tpr_resid_from_one=True`` by default.
        """
        with openany(self.filename, mode="rb") as infile:
            tprf = infile.read()
        data = tpr_utils.TPXUnpacker(tprf)
        try:
            th = tpr_utils.read_tpxheader(data)  # tpxheader
        except (EOFError, ValueError):
            msg = f"{self.filename}: Invalid tpr file or cannot be recognized"
            logger.critical(msg)
            raise IOError(msg)

        self._log_header(th)

        # Starting with gromacs 2020 (tpx version 119), the body of the file
        # is encoded differently. We change the unpacker accordingly.
        if th.fver >= S.tpxv_AddSizeField and th.fgen >= 27:
            actual_body_size = len(data.get_buffer()) - data.get_position()
            if actual_body_size == 4 * th.sizeOfTprBody:
                # See issue #2428.
                msg = (
                    "TPR files produced with beta versions of gromacs 2020 "
                    "are not supported."
                )
                logger.critical(msg)
                raise IOError(msg)
            data = tpr_utils.TPXUnpacker2020.from_unpacker(data)

        state_ngtc = th.ngtc  # done init_state() in src/gmxlib/tpxio.c
        if th.bBox:
            tpr_utils.extract_box_info(data, th.fver)

        if state_ngtc > 0:
            if th.fver < 69:  # redundancy due to  different versions
                tpr_utils.ndo_real(data, state_ngtc)
            tpr_utils.ndo_real(
                data, state_ngtc
            )  # relevant to Berendsen tcoupl_lambda

        if th.bTop:
            tpr_top = tpr_utils.do_mtop(
                data,
                th.fver,
                tpr_resid_from_one=tpr_resid_from_one,
                precision=th.precision,
            )
        else:
            msg = f"{self.filename}: No topology found in tpr file"
            logger.critical(msg)
            raise IOError(msg)

        tpr_top.add_TopologyAttr(Resnums(tpr_top.resids.values.copy()))

        return tpr_top

    def _log_header(self, th):
        logger.info(f"Gromacs version   : {th.ver_str}")
        logger.info(f"tpx version       : {th.fver}")
        logger.info(f"tpx generation    : {th.fgen}")
        logger.info(f"tpx precision     : {th.precision}")
        logger.info(f"tpx file_tag      : {th.file_tag}")
        logger.info(f"tpx natoms        : {th.natoms}")
        logger.info(f"tpx ngtc          : {th.ngtc}")
        logger.info(f"tpx fep_state     : {th.fep_state}")
        logger.info(f"tpx lambda        : {th.lamb}")
        logger.debug(f"tpx bIr (input record): {th.bIr}")
        logger.debug(f"tpx bTop         : {th.bTop}")
        logger.debug(f"tpx bX           : {th.bX}")
        logger.debug(f"tpx bV           : {th.bV}")
        logger.debug(f"tpx bF           : {th.bF}")
        logger.debug(f"tpx bBox         : {th.bBox}")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
from numpy.testing import assert_equal, assert_allclose
import numpy as np
from io import StringIO

import MDAnalysis as mda
from MDAnalysisTests.topology.base import ParserBase
from MDAnalysis.tests.datafiles import (
    LAMMPSdata,
    LAMMPScnt,
    LAMMPScnt2,
    LAMMPShyd,
    LAMMPShyd2,
    LAMMPSdata_deletedatoms,
    LAMMPSDUMP,
    LAMMPSDUMP_long,
    LAMMPSDUMP_allinfo,
    LAMMPSDUMP_nomass_elemx,
    LAMMPSdata_PairIJ,
)


class LammpsBase(ParserBase):
    parser = mda.topology.LAMMPSParser.DATAParser
    expected_n_segments = 1
    expected_attrs = [
        "types",
        "resids",
        "masses",
        "charges",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
    ]

    def test_n_atom_types(self, top):
        assert_equal(len(set(top.types.values)), self.expected_n_atom_types)

    def test_n_bonds(self, top):
        if self.ref_n_bonds:
            assert_equal(len(top.bonds.values), self.ref_n_bonds)
        else:
            assert not hasattr(top, "bonds")

    def test_bond_member(self, top):
        if self.ref_n_bonds:
            assert self.ref_bond in top.bonds.values

    def test_n_angles(self, top):
        if self.ref_n_angles:
            assert_equal(len(top.angles.values), self.ref_n_angles)
        else:
            assert not hasattr(self.top, "angles")

    def test_angle_member(self, top):
        if self.ref_n_angles:
            assert self.ref_angle in top.angles.values

    def test_n_dihedrals(self, top):
        if self.ref_n_dihedrals:
            assert_equal(len(top.dihedrals.values), self.ref_n_dihedrals)
        else:
            assert not hasattr(self.top, "dihedrals")

    def test_dihedral_member(self, top):
        if self.ref_n_dihedrals:
            assert self.ref_dihedral in top.dihedrals.values

    def test_n_impropers(self, top):
        if self.ref_n_impropers:
            assert_equal(len(top.impropers.values), self.ref_n_impropers)
        else:
            assert not hasattr(self.top, "impropers")

    def test_improper_member(self, top):
        if self.ref_n_impropers:
            assert self.ref_improper in top.impropers.values

    def test_creates_universe(self, filename):
        u = mda.Universe(filename, format="DATA")

    def test_guessed_attributes(self, filename):
        u = mda.Universe(filename, format="DATA")
        for attr in self.guessed_attrs:
            assert hasattr(u.atoms, attr)


class TestLammpsData(LammpsBase):
    """Tests the reading of lammps .data topology files.

    The reading of coords and velocities is done separately in
    test_coordinates
    """

    ref_filename = LAMMPSdata
    expected_n_atoms = 18364
    expected_n_atom_types = 10
    expected_n_residues = 25
    ref_n_bonds = 18336
    ref_bond = (12, 14)
    ref_n_angles = 29904
    ref_angle = (3, 6, 9)
    ref_n_dihedrals = 5712
    ref_dihedral = (82, 85, 88, 89)
    ref_n_impropers = 0


class TestLAMMPSCNT(LammpsBase):
    expected_n_atoms = 604
    expected_n_atom_types = 1
    expected_n_residues = 1
    ref_n_bonds = 906
    ref_bond = (9, 467)
    ref_n_angles = 1812
    ref_angle = (17, 16, 31)
    ref_n_dihedrals = 3624
    ref_dihedral = (22, 39, 40, 41)
    ref_n_impropers = 604
    ref_improper = (210, 159, 212, 566)

    @pytest.fixture(params=[LAMMPScnt, LAMMPScnt2])
    def filename(self, request):
        return request.param


class TestLAMMPSHYD(LammpsBase):
    expected_n_atoms = 2
    expected_n_atom_types = 1
    expected_n_residues = 1
    ref_n_bonds = 1
    ref_bond = (0, 1)
    ref_n_angles = 0
    ref_n_dihedrals = 0
    ref_n_impropers = 0

    @pytest.fixture(params=[LAMMPShyd, LAMMPShyd2])
    def filename(self, request):
        return request.param


class TestLAMMPSDeletedAtoms(LammpsBase):
    ref_filename = LAMMPSdata_deletedatoms
    expected_n_atoms = 10
    expected_n_atom_types = 2
    expected_n_residues = 1

    ref_n_bonds = 9
    ref_bond = (0, 3)
    ref_n_angles = 0
    ref_n_dihedrals = 0
    ref_n_impropers = 0

    def test_atom_ids(self, filename):
        u = mda.Universe(filename)

        assert_equal(
            u.atoms.ids,
            [1, 10, 1002, 2003, 2004, 2005, 2006, 2007, 2008, 2009],
        )

    def test_traj(self, filename):
        u = mda.Universe(filename)

        assert_equal(
            u.atoms.positions,
            np.array(
                [
                    [11.8998565674, 48.4455718994, 19.0971984863],
                    [14.5285415649, 50.6892776489, 19.9419136047],
                    [12.8466796875, 48.1473007202, 18.6461906433],
                    [11.0093536377, 48.7145767212, 18.5247917175],
                    [12.4033203125, 49.2582168579, 20.2825050354],
                    [13.0947723389, 48.8437194824, 21.0175533295],
                    [11.540184021, 49.6138534546, 20.8459072113],
                    [13.0085144043, 50.6062469482, 19.9141769409],
                    [12.9834518433, 51.1562423706, 18.9713554382],
                    [12.6588821411, 51.4160842896, 20.5548400879],
                ],
                dtype=np.float32,
            ),
        )


class TestLammpsDataPairIJ(LammpsBase):
    """Tests the reading of lammps .data topology file with a
    PairIJ Coeffs section
    """

    expected_attrs = [
        "types",
        "resids",
        "masses",
        "bonds",
        "angles",
        "dihedrals",
        "impropers",
    ]
    ref_filename = LAMMPSdata_PairIJ
    expected_n_atoms = 800
    expected_n_atom_types = 2
    expected_n_residues = 1
    ref_n_bonds = 799
    ref_bond = (397, 398)
    ref_n_angles = 390
    ref_angle = (722, 723, 724)
    ref_n_dihedrals = 385
    ref_dihedral = (722, 723, 724, 725)
    ref_n_impropers = 0


LAMMPS_NORESID = """\
LAMMPS data file via write_data, version 11 Aug 2017, timestep = 0

1 atoms
1 atom types

0.0000000000000000e+00 4.3008000000000003e+01 xlo xhi
0.0000000000000000e+00 4.3008000000000003e+01 ylo yhi
0.0000000000000000e+00 4.3008000000000003e+01 zlo zhi

Masses

1 28

Pair Coeffs # lj/cut

1 0.1892 3.75

Atoms # atomic

1 1 3.7151744275286681e+01 1.8684434743140471e+01 1.9285127961842125e+01 0 0 0
"""


def test_noresid():
    u = mda.Universe(
        StringIO(LAMMPS_NORESID), format="data", atom_style="id type x y z"
    )
    assert len(u.atoms) == 1

    assert_equal(u.atoms[0].mass, 28.0)
    assert_equal(
        u.atoms.positions,
        np.array(
            [
                [
                    3.7151744275286681e01,
                    1.8684434743140471e01,
                    1.9285127961842125e01,
                ]
            ],
            dtype=np.float32,
        ),
    )


def test_noresid_failure():
    with pytest.raises(
        ValueError,
        match=".+?You can supply a description of the atom_style.+?",
    ):
        u = mda.Universe(StringIO(LAMMPS_NORESID), format="data")


def test_interpret_atom_style():
    style = mda.topology.LAMMPSParser.DATAParser._interpret_atom_style(
        "id charge type z y x"
    )

    assert isinstance(style, dict)
    assert style["id"] == 0
    assert style["type"] == 2
    assert style["charge"] == 1
    assert style["x"] == 5
    assert style["y"] == 4
    assert style["z"] == 3


def test_interpret_atom_style_missing():
    with pytest.raises(
        ValueError, match="atom_style string missing required.+?"
    ):
        style = mda.topology.LAMMPSParser.DATAParser._interpret_atom_style(
            "id charge z y x"
        )


class LammpsDumpBase(ParserBase):
    """Tests the reading of lammps dump files for topology information."""

    expected_attrs = ["types", "masses"]
    expected_n_segments = 1

    parser = mda.topology.LAMMPSParser.LammpsDumpParser

    def test_creates_universe(self):
        u = mda.Universe(self.ref_filename, format="LAMMPSDUMP")

        assert isinstance(u, mda.Universe)
        assert len(u.atoms) == self.expected_n_atoms

    def test_masses(self, top):
        assert_allclose(top.masses.values[:7], self.expected_masses)

    def test_guessed_attributes(self, filename):
        u = mda.Universe(filename, format="LAMMPSDUMP")
        for attr in self.guessed_attrs:
            assert hasattr(u.atoms, attr)

    def test_id_ordering(self, top):
        # ids are nonsequential in file, but should get rearranged
        # Check that the ids are in order
        assert_equal(top.ids.values, np.arange(1, self.expected_n_atoms + 1))

    def test_types(self, top):
        assert (top.types.values[:7] == self.expected_types).all()

    def test_elements(self, top):
        if self.expected_elements:
            assert (top.elements.values[:7] == self.expected_elements).all()
        else:
            assert not hasattr(top, "elements")

    def test_charges(self, top):
        if self.expected_charges:
            assert_allclose(top.charges.values[:7], self.expected_charges)
        else:
            assert not hasattr(top, "charges")

    def test_resids(self, top):
        assert_equal(top.resids.values, self.expected_resids)


# Test that the parser works for simple dump files with no mass information
class TestDumpParserBasic(LammpsDumpBase):
    expected_n_atoms = 24
    expected_n_residues = 1
    expected_resids = [1]
    expected_masses = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    expected_types = ["2", "1", "1", "2", "1", "1", "2"]
    # 0 denotes that charges are not to be expected
    expected_charges = 0
    expected_elements = 0
    ref_filename = LAMMPSDUMP

    def test_masses_warning(self):
        # masses are mandatory, but badly guessed
        # check that user is alerted
        with self.parser(self.ref_filename) as p:
            with pytest.warns(UserWarning, match="Guessed all Masses to 1.0"):
                p.parse()


# this tests that topology can still be constructed if non-standard or uneven
# column present.
class TestDumpParserLong(LammpsDumpBase):
    expected_n_atoms = 24
    expected_n_residues = 1
    expected_resids = [1]
    expected_masses = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    expected_types = ["2", "1", "1", "2", "1", "1", "2"]
    expected_charges = 0
    expected_elements = 0
    ref_filename = LAMMPSDUMP_long

    def test_masses_warning(self):
        # masses are mandatory, but badly guessed
        # check that user is alerted
        with self.parser(self.ref_filename) as p:
            with pytest.warns(UserWarning, match="Guessed all Masses to 1.0"):
                p.parse()


class TestDumpParserFull(LammpsDumpBase):
    ref_filename = LAMMPSDUMP_allinfo
    expected_attrs = ["types", "masses", "charges", "elements"]
    expected_n_atoms = 30
    expected_n_residues = 3
    expected_resids = [1, 2, 3]
    expected_masses = [12.011, 1.008, 1.008, 1.008, 12.011, 15.9994, 14.007]
    expected_types = ["9", "6", "6", "6", "7", "15", "14"]
    expected_charges = [-0.27, 0.09, 0.09, 0.09, 0.51, -0.51, -0.47]
    expected_elements = ["C", "H", "H", "H", "C", "O", "N"]


# Test whether mass information is guessed correctly
# when element information is available
# and whether the user is warned about unknown elements
class TestDumpParserElementX(LammpsDumpBase):
    ref_filename = LAMMPSDUMP_nomass_elemx
    expected_attrs = ["types", "masses", "charges", "elements"]
    expected_n_atoms = 5
    expected_n_residues = 1
    expected_resids = [1]
    expected_masses = [12.011, 1, 1.008, 1.008, 12.011]
    expected_types = ["9", "6", "6", "6", "7"]
    expected_charges = [-0.27, 0.09, 0.09, 0.09, 0.51]
    expected_elements = ["C", "", "H", "H", "C"]

    def test_elementwarn(self):
        with self.parser(self.ref_filename) as p:
            with pytest.warns(
                UserWarning,
                match="Unknown element X found for some atoms. "
                "These have been given an empty element record.",
            ):
                p.parse()

    def test_masses_warning(self):
        # masses are mandatory, but badly guessed
        # check that user is alerted
        with self.parser(self.ref_filename) as p:
            with pytest.warns(
                UserWarning,
                match="No mass column found in dump file. "
                "Using guessed masses from element info.",
            ):
                p.parse()


LAMMPS_DUMP_NOATOMIDS = """\
ITEM: TIMESTEP
0
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS pp pp pp
-1.6350000000000001e-01 3.8836500000000001e+01
-1.7050000000000001e-01 3.8829500000000003e+01
-1.1550000000000001e-01 9.4884500000000003e+01
ITEM: ATOMS mol type x y z q mass
1 9 6.10782 11.4384 0.798271 -0.27 12.011
1 6 5.73938 10.7721 1.60738 0.09 1.008
"""


def test_noatomid_failure():
    with pytest.raises(ValueError, match="No id column found in dump file"):
        mda.Universe(StringIO(LAMMPS_DUMP_NOATOMIDS), format="LAMMPSDUMP")


LAMMPS_DUMP_NOTYPES = """\
ITEM: TIMESTEP
0
ITEM: NUMBER OF ATOMS
2
ITEM: BOX BOUNDS pp pp pp
-1.6350000000000001e-01 3.8836500000000001e+01
-1.7050000000000001e-01 3.8829500000000003e+01
-1.1550000000000001e-01 9.4884500000000003e+01
ITEM: ATOMS id mol x y z q mass
1 1 6.10782 11.4384 0.798271 -0.27 12.011
2 1 5.73938 10.7721 1.60738 0.09 1.008
"""


def test_notypes():
    u = mda.Universe(StringIO(LAMMPS_DUMP_NOTYPES), format="LAMMPSDUMP")
    assert len(u.atoms) == 2
    assert_equal(u.atoms.types, np.ones(2, dtype=object))


def test_notypes_warn():
    with pytest.warns(UserWarning, match="Set all atom types to 1"):
        mda.Universe(StringIO(LAMMPS_DUMP_NOTYPES), format="LAMMPSDUMP")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_allclose
import MDAnalysis
from MDAnalysis.visualization import streamlines, streamlines_3D
from MDAnalysis.coordinates.XTC import XTCWriter
from MDAnalysisTests.datafiles import Martini_membrane_gro
import pytest
from pytest import approx
import matplotlib.pyplot as plt
import os


@pytest.fixture(scope="session")
def univ():
    u = MDAnalysis.Universe(Martini_membrane_gro)
    return u


@pytest.fixture(scope="session")
def membrane_xtc(tmpdir_factory, univ):
    x_delta, y_delta, z_delta = 0.5, 0.3, 0.2
    tmp_xtc = tmpdir_factory.mktemp("streamlines").join("dummy.xtc")

    with XTCWriter(str(tmp_xtc), n_atoms=univ.atoms.n_atoms) as xtc_writer:
        for i in range(5):
            univ.atoms.translate([x_delta, y_delta, z_delta])
            xtc_writer.write(univ.atoms)
            x_delta += 0.1
            y_delta += 0.08
            z_delta += 0.02
    return str(tmp_xtc)


def test_produce_list_indices_point_in_polygon_this_frame():
    # Define two squares:
    # square1 covers the area [0,1]x[0,1]
    # square2 covers the area [2,3]x[2,3]
    square1 = [(0, 0), (1, 0), (1, 1), (0, 1)]
    square2 = [(2, 2), (3, 2), (3, 3), (2, 3)]

    # Create a list of vertex coordinates (for two squares)
    vertex_list = [square1, square2]

    # Define points:
    # Point [0.5, 0.5] lies inside square1.
    # Point [1.5, 1.5] lies outside both squares.
    # Point [2.5, 2.5] lies inside square2.
    # Point [3.5, 3.5] lies outside both squares.
    points = np.array([[0.5, 0.5], [1.5, 1.5], [2.5, 2.5], [3.5, 3.5]])

    # Call the function under test.
    result = streamlines._produce_list_indices_point_in_polygon_this_frame(
        vertex_list, points
    )

    # np.where returns a tuple; thus for each square we expect:
    # For square1: point index 0 is inside → (array([0]),)
    # For square2: point index 2 is inside → (array([2]),)
    expected = [(np.array([0]),), (np.array([2]),)]

    # Check that each result matches the expected indices.
    for res_tuple, exp_tuple in zip(result, expected):
        np.testing.assert_array_equal(res_tuple[0], exp_tuple[0])


def test_produce_list_centroids_empty():
    # Simulate an empty index set for one square:
    list_indices = [(np.array([]),)]
    # Dummy particle coordinate array (won't be used since indices is empty)
    pts = np.array([[0, 0], [1, 1]])
    result = streamlines._produce_list_centroids_this_frame(list_indices, pts)
    assert result == [None]


def test_produce_list_centroids_single_square():
    # Create an array of particle coordinates
    pts = np.array([[0, 0], [1, 1], [2, 2], [3, 3]])
    # Choose indices that pick points [1] and [3]
    indices_tuple = (np.array([1, 3]),)
    list_indices = [indices_tuple]
    result = streamlines._produce_list_centroids_this_frame(list_indices, pts)
    expected = np.array([2.0, 2.0])
    np.testing.assert_allclose(result[0], expected)


def test_produce_list_centroids_multiple_squares():
    pts = np.array([[0, 0], [2, 2], [4, 4], [6, 6]])
    # First square will use pts[0] and pts[2] -> average is [2,2]
    indices1 = (np.array([0, 2]),)
    # Second square will use pts[1] and pts[3] -> average is [4,4]
    indices2 = (np.array([1, 3]),)
    list_indices = [indices1, indices2]
    result = streamlines._produce_list_centroids_this_frame(list_indices, pts)
    expected1 = np.array([2, 2])
    expected2 = np.array([4, 4])
    np.testing.assert_array_equal(result[0], expected1)
    np.testing.assert_array_equal(result[1], expected2)


def test_adjacent_squares():
    # Test two adjacent squares that share a boundary.
    # Square1 covers [0,1]x[0,1] and square2 covers [1,2]x[0,1].
    # A point at [0.5, 0.5] should be in square1 and one at [1.5,0.5] should be in square2.
    square1 = [(0, 0), (1, 0), (1, 1), (0, 1)]
    square2 = [(1, 0), (2, 0), (2, 1), (1, 1)]
    vertex_list = [square1, square2]
    points = np.array(
        [
            [0.5, 0.5],
            [1.5, 0.5],
        ]
    )
    result = streamlines._produce_list_indices_point_in_polygon_this_frame(
        vertex_list, points
    )
    expected = [(np.array([0]),), (np.array([1]),)]
    for res_tuple, exp_tuple in zip(result, expected):
        np.testing.assert_array_equal(res_tuple[0], exp_tuple[0])


def test_point_on_boundary():
    # Test that a point exactly on the square's boundary is not considered inside.
    # For a square covering [0,1]x[0,1], a point at [1,0.5] lies exactly on the right edge.
    # By default, matplotlib.path.Path.contains_points includes boundary points.
    square = [(0, 0), (1, 0), (1, 1), (0, 1)]
    vertex_list = [square]
    points = np.array([[1, 0.5]])  # exactly on the boundary
    result = streamlines._produce_list_indices_point_in_polygon_this_frame(
        vertex_list, points
    )
    expected = [(np.array([0], dtype=int),)]
    np.testing.assert_array_equal(result, expected)


def test_points_on_boundary_of_two_adjacent_squares():
    square1 = [(0, 0), (1, 0), (1, 1), (0, 1)]
    square2 = [(1, 0), (2, 0), (2, 1), (1, 1)]
    vertex_list = [square1, square2]
    points = np.array([[1, 0.5], [1, 0.7]])  # exactly on the boundary
    result = streamlines._produce_list_indices_point_in_polygon_this_frame(
        vertex_list, points
    )
    expected = [(np.array([0, 1], dtype=int),), (np.array([0, 1], dtype=int),)]
    np.testing.assert_array_equal(result, expected)


def test_per_core_work_2D(membrane_xtc, univ):
    xmin = univ.atoms.positions[..., 0].min()
    xmax = univ.atoms.positions[..., 0].max()
    ymin = univ.atoms.positions[..., 1].min()
    ymax = univ.atoms.positions[..., 1].max()
    tuple_of_limits = (xmin, xmax, ymin, ymax)
    grid = streamlines.produce_grid(
        tuple_of_limits=tuple_of_limits, grid_spacing=20
    )
    (
        list_square_vertex_arrays_per_core,
        list_parent_index_values,
        _,
        _,
    ) = streamlines.split_grid(grid=grid, num_cores=1)
    values = streamlines.per_core_work(
        topology_file_path=Martini_membrane_gro,
        trajectory_file_path=membrane_xtc,
        list_square_vertex_arrays_this_core=list_square_vertex_arrays_per_core[
            0
        ],
        MDA_selection="name PO4",
        start_frame=1,
        end_frame=2,
        reconstruction_index_list=list_parent_index_values[0],
        maximum_delta_magnitude=2.0,
    )
    for entry in values:
        res = entry[1]
        np.testing.assert_allclose(res[:2], np.array([0.8, 0.5]), atol=1e-1)


def test_streamplot_2D(membrane_xtc, univ):
    # regression test the data structures
    # generated by the 2D streamplot code
    u1, v1, avg, std = streamlines.generate_streamlines(
        topology_file_path=Martini_membrane_gro,
        trajectory_file_path=membrane_xtc,
        grid_spacing=20,
        MDA_selection="name PO4",
        start_frame=1,
        end_frame=2,
        xmin=univ.atoms.positions[..., 0].min(),
        xmax=univ.atoms.positions[..., 0].max(),
        ymin=univ.atoms.positions[..., 1].min(),
        ymax=univ.atoms.positions[..., 1].max(),
        maximum_delta_magnitude=2.0,
        num_cores=1,
    )
    assert_allclose(
        u1,
        np.array(
            [
                [0.79999924, 0.79999924, 0.80000687, 0.79999542, 0.79998779],
                [0.80000019, 0.79999542, 0.79999924, 0.79999542, 0.80001068],
                [0.8000021, 0.79999924, 0.80001068, 0.80000305, 0.79999542],
                [0.80000019, 0.79999542, 0.80001068, 0.80000305, 0.80000305],
                [0.79999828, 0.80000305, 0.80000305, 0.80000305, 0.79999542],
            ]
        ),
    )
    assert_allclose(
        v1,
        np.array(
            [
                [0.53999901, 0.53999996, 0.53999996, 0.53999996, 0.54000092],
                [0.5399971, 0.54000092, 0.54000092, 0.54000092, 0.5399971],
                [0.54000473, 0.54000473, 0.54000092, 0.5399971, 0.54000473],
                [0.54000092, 0.53999329, 0.53999329, 0.53999329, 0.54000092],
                [0.54000092, 0.53999329, 0.53999329, 0.54000092, 0.53999329],
            ]
        ),
    )
    assert avg == pytest.approx(0.965194167)
    assert std == pytest.approx(4.444808820e-06)


def test_streamplot_2D_zero_return(membrane_xtc, univ, tmpdir):
    # simple roundtrip test to ensure that
    # zeroed arrays are returned by the 2D streamplot
    # code when called with an empty selection
    u1, v1, avg, std = streamlines.generate_streamlines(
        topology_file_path=Martini_membrane_gro,
        trajectory_file_path=membrane_xtc,
        grid_spacing=20,
        MDA_selection="name POX",
        start_frame=1,
        end_frame=2,
        xmin=univ.atoms.positions[..., 0].min(),
        xmax=univ.atoms.positions[..., 0].max(),
        ymin=univ.atoms.positions[..., 1].min(),
        ymax=univ.atoms.positions[..., 1].max(),
        maximum_delta_magnitude=2.0,
        num_cores=1,
    )
    assert_allclose(u1, np.zeros((5, 5)))
    assert_allclose(v1, np.zeros((5, 5)))
    assert avg == approx(0.0)
    assert std == approx(0.0)


def test_streamplot_2D_dual_core(membrane_xtc, univ, tmpdir):
    # simple test to ensure that it runs with multiple cores
    u1, v1, avg, std = streamlines.generate_streamlines(
        topology_file_path=Martini_membrane_gro,
        trajectory_file_path=membrane_xtc,
        grid_spacing=20,
        MDA_selection="name POX",
        start_frame=1,
        end_frame=2,
        xmin=univ.atoms.positions[..., 0].min(),
        xmax=univ.atoms.positions[..., 0].max(),
        ymin=univ.atoms.positions[..., 1].min(),
        ymax=univ.atoms.positions[..., 1].max(),
        maximum_delta_magnitude=2.0,
        num_cores=2,
    )
    assert_allclose(u1, np.zeros((5, 5)))
    assert_allclose(v1, np.zeros((5, 5)))
    assert avg == approx(0.0)
    assert std == approx(0.0)


def test_streamplot_3D(membrane_xtc, univ, tmpdir):
    # because mayavi is too heavy of a dependency
    # for a roundtrip plotting test, simply
    # aim to check for sensible values
    # returned by generate_streamlines_3d
    dx, dy, dz = streamlines_3D.generate_streamlines_3d(
        topology_file_path=Martini_membrane_gro,
        trajectory_file_path=membrane_xtc,
        grid_spacing=20,
        MDA_selection="name PO4",
        start_frame=1,
        end_frame=2,
        xmin=univ.atoms.positions[..., 0].min(),
        xmax=univ.atoms.positions[..., 0].max(),
        ymin=univ.atoms.positions[..., 1].min(),
        ymax=univ.atoms.positions[..., 1].max(),
        zmin=univ.atoms.positions[..., 2].min(),
        zmax=univ.atoms.positions[..., 2].max(),
        maximum_delta_magnitude=2.0,
        num_cores=1,
    )
    assert dx.shape == (5, 5, 2)
    assert dy.shape == (5, 5, 2)
    assert dz.shape == (5, 5, 2)
    assert dx[4, 4, 0] == approx(0.700004, abs=1e-5)
    assert dy[0, 0, 0] == approx(0.460000, abs=1e-5)
    assert dz[2, 2, 0] == approx(0.240005, abs=1e-5)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import numpy as np
from numpy.testing import assert_equal

from MDAnalysis.lib._cutil import (
    unique_int_1d,
    find_fragments,
    _in2d,
)


@pytest.mark.parametrize(
    "values",
    (
        [],  # empty array
        [1, 1, 1, 1],  # all identical
        [2, 3, 5, 7],  # all different, monotonic
        [5, 2, 7, 3],  # all different, non-monotonic
        [1, 2, 2, 4, 4, 6],  # duplicates, monotonic
        [1, 2, 2, 6, 4, 4],  # duplicates, non-monotonic
    ),
)
def test_unique_int_1d(values):
    array = np.array(values, dtype=np.intp)
    ref = np.unique(array)
    res = unique_int_1d(array)
    assert_equal(res, ref)
    assert type(res) == type(ref)
    assert res.dtype == ref.dtype


@pytest.mark.parametrize(
    "edges,ref",
    [
        ([[0, 1], [1, 2], [2, 3], [3, 4]], [[0, 1, 2, 3, 4]]),  # linear chain
        (
            [[0, 1], [1, 2], [2, 3], [3, 4], [4, 10]],
            [[0, 1, 2, 3, 4]],
        ),  # unused edge (4, 10)
        ([[0, 1], [1, 2], [2, 3]], [[0, 1, 2, 3], [4]]),  # lone atom
        (
            [[0, 1], [1, 2], [2, 0], [3, 4], [4, 3]],
            [[0, 1, 2], [3, 4]],
        ),  # circular
    ],
)
def test_find_fragments(edges, ref):
    atoms = np.arange(5)

    fragments = find_fragments(atoms, edges)

    assert len(fragments) == len(ref)
    for frag, r in zip(fragments, ref):
        assert_equal(frag, r)


def test_in2d():
    arr1 = np.array([[1, 2], [3, 4], [5, 6]], dtype=np.intp)
    arr2 = np.array([[3, 4], [2, 1], [5, 5], [6, 6]], dtype=np.intp)

    result = _in2d(arr1, arr2)

    assert_equal(result, np.array([False, True, False]))


@pytest.mark.parametrize(
    "arr1,arr2",
    [
        (
            np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.intp),
            np.array([[1, 2], [3, 4]], dtype=np.intp),
        ),
        (
            np.array([[1, 2], [3, 4]], dtype=np.intp),
            np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.intp),
        ),
    ],
)
def test_in2d_VE(arr1, arr2):
    with pytest.raises(
        ValueError, match=r"Both arrays must be \(n, 2\) arrays"
    ):
        _in2d(arr1, arr2)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import multiprocessing

import numpy as np
import os
import shutil
import pytest
import pickle
from numpy.testing import assert_equal

import MDAnalysis as mda
import MDAnalysis.coordinates
from MDAnalysis.coordinates.core import get_reader_for
from MDAnalysis.coordinates.GSD import HAS_GSD
from MDAnalysis.analysis.rms import RMSD

from MDAnalysisTests.datafiles import (
    CRD,
    PSF,
    DCD,
    DMS,
    DLP_CONFIG,
    DLP_HISTORY,
    FHIAIMS,
    INPCRD,
    GMS_ASYMOPT,
    GMS_SYMOPT,
    GRO,
    GSD,
    GSD_long,
    LAMMPSdata_mini,
    LAMMPSDUMP,
    mol2_molecules,
    MMTF,
    NCDF,
    PDB,
    PDB_small,
    PDB_multiframe,
    PDBQT_input,
    PQR,
    TRC_PDB_VAC,
    TRC_TRAJ1_VAC,
    TRC_TRAJ2_VAC,
    TRR,
    TRJ,
    TRZ,
    TXYZ,
    XTC,
    XPDB_small,
    XYZ_mini,
    XYZ,
    XYZ_bz2,
)


@pytest.fixture(
    params=[
        (PSF, DCD),
        (GRO, XTC),
        (PDB_multiframe,),
        (XYZ,),
        (XYZ_bz2,),  # .bz2
        (GMS_SYMOPT,),  # .gms
        (GMS_ASYMOPT,),  # .gz
        pytest.param(
            (GSD_long,),
            marks=pytest.mark.skipif(not HAS_GSD, reason="gsd not installed"),
        ),
        (NCDF,),
        (np.arange(150).reshape(5, 10, 3).astype(np.float64),),
        (GRO, [GRO, GRO, GRO, GRO, GRO]),
        (PDB, [PDB, PDB, PDB, PDB, PDB]),
        (GRO, [XTC, XTC]),
        (TRC_PDB_VAC, TRC_TRAJ1_VAC),
        (TRC_PDB_VAC, [TRC_TRAJ1_VAC, TRC_TRAJ2_VAC]),
    ]
)
def u(request):
    if len(request.param) == 1:
        f = request.param[0]
        return mda.Universe(f, to_guess=())
    else:
        top, trj = request.param
        return mda.Universe(top, trj)


@pytest.fixture(scope="function")
def temp_xtc(tmp_path):
    fresh_xtc = tmp_path / "testing.xtc"
    shutil.copy(XTC, fresh_xtc)
    # In principle there is no possibility that the offset for
    # fresh_xtc exists as it is a fresh copy for every test. However,
    # the code is left for documentation.
    try:
        os.remove(MDAnalysis.coordinates.XDR.offsets_filename(fresh_xtc))
    except OSError:
        pass
    return fresh_xtc


# Define target functions here
# inside test functions doesn't work
def cog(u, ag, frame_id):
    u.trajectory[frame_id]

    return ag.center_of_geometry()


def test_multiprocess_COG(u):
    ag = u.atoms[2:5]

    ref = np.array([cog(u, ag, i) for i in range(3)])

    p = multiprocessing.Pool(2)
    res = np.array([p.apply(cog, args=(u, ag, i)) for i in range(3)])
    p.close()
    assert_equal(ref, res)


def getnames(u, ix):
    # Check topology stuff works
    return u.atoms[ix].name


def test_universe_unpickle_in_new_process():
    u = mda.Universe(GRO, XTC)
    ref = [getnames(u, i) for i in range(3)]

    p = multiprocessing.Pool(2)
    res = [p.apply(getnames, args=(u, i)) for i in range(3)]
    p.close()

    assert_equal(ref, res)


def test_creating_multiple_universe_without_offset(temp_xtc, ncopies=3):
    #  test if they can be created without generating
    #  the offset simultaneously.
    #  The tested XTC file is way too short to induce a race scenario but the
    #  test is included as documentation for the scenario that used to create
    #  a problem (see PR #3375 and issues #3230, #1988)

    args = (GRO, str(temp_xtc))
    with multiprocessing.Pool(2) as p:
        universes = [p.apply_async(mda.Universe, args) for i in range(ncopies)]
        universes = [universe.get() for universe in universes]

    assert_equal(
        universes[0].trajectory._xdr.offsets,
        universes[1].trajectory._xdr.offsets,
    )


@pytest.fixture(
    params=[
        # formatname, filename
        ("CRD", CRD, dict()),
        ("DATA", LAMMPSdata_mini, dict(n_atoms=1)),
        ("DCD", DCD, dict()),
        ("DMS", DMS, dict()),
        ("CONFIG", DLP_CONFIG, dict()),
        ("FHIAIMS", FHIAIMS, dict()),
        ("HISTORY", DLP_HISTORY, dict()),
        ("INPCRD", INPCRD, dict()),
        ("LAMMPSDUMP", LAMMPSDUMP, dict()),
        ("GMS", GMS_ASYMOPT, dict()),
        ("GRO", GRO, dict()),
        pytest.param(
            ("GSD", GSD, dict()),
            marks=pytest.mark.skipif(not HAS_GSD, reason="gsd not installed"),
        ),
        ("MMTF", MMTF, dict()),
        ("MOL2", mol2_molecules, dict()),
        ("PDB", PDB_small, dict()),
        ("PQR", PQR, dict()),
        ("PDBQT", PDBQT_input, dict()),
        ("TRR", TRR, dict()),
        ("TRZ", TRZ, dict(n_atoms=8184)),
        ("TRJ", TRJ, dict(n_atoms=252)),
        ("XTC", XTC, dict()),
        ("XPDB", XPDB_small, dict()),
        ("XYZ", XYZ_mini, dict()),
        ("NCDF", NCDF, dict()),
        ("TXYZ", TXYZ, dict()),
        ("memory", np.arange(60).reshape(2, 10, 3).astype(np.float64), dict()),
        ("TRC", TRC_TRAJ1_VAC, dict()),
        ("CHAIN", [TRC_TRAJ1_VAC, TRC_TRAJ2_VAC], dict()),
        ("CHAIN", [GRO, GRO, GRO], dict()),
        ("CHAIN", [PDB, PDB, PDB], dict()),
        ("CHAIN", [XTC, XTC, XTC], dict()),
    ]
)
def ref_reader(request):
    fmt_name, filename, extras = request.param
    r = get_reader_for(filename, format=fmt_name)(filename, **extras)
    try:
        yield r
    finally:
        # make sure file handle is closed afterwards
        r.close()


def test_readers_pickle(ref_reader):
    ps = pickle.dumps(ref_reader)
    reanimated = pickle.loads(ps)
    assert len(ref_reader) == len(reanimated)
    try:
        ref_reader[2]
        ref_reader[0]
    except IndexError:
        # single frame files
        pass
    assert_equal(reanimated.ts, ref_reader.ts)


def test_analysis_pickle():
    u = mda.Universe(PSF, DCD)
    rmsd = RMSD(u.atoms, u.atoms)
    rmsd.run()
    rmsd_p = pickle.dumps(rmsd)
    rmsd_new = pickle.loads(rmsd_p)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
import logging
import pytest
import copy

from numpy.testing import (
    assert_allclose,
    assert_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_almost_equal,
)

import MDAnalysis
from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import (
    HydrogenBondAnalysis,
)
from MDAnalysis.exceptions import NoDataError
from MDAnalysisTests.datafiles import waterPSF, waterDCD


class TestHydrogenBondAnalysisTIP3P(object):

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        "donors_sel": "name OH2",
        "hydrogens_sel": "name H1 H2",
        "acceptors_sel": "name OH2",
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @pytest.fixture(scope="class")
    def h(self, universe, client_HydrogenBondAnalysis):
        h = HydrogenBondAnalysis(universe, **self.kwargs)
        h.run(**client_HydrogenBondAnalysis)
        return h

    @pytest.fixture(scope="class")
    def h_modified(self, universe, client_HydrogenBondAnalysis):
        kwargs = self.kwargs.copy()
        donors_sel = kwargs.pop("donors_sel")
        hydrogens_sel = kwargs.pop("hydrogens_sel")
        acceptors_sel = kwargs.pop("acceptors_sel")
        h = HydrogenBondAnalysis(universe, **kwargs)
        h.donors_sel = donors_sel
        h.hydrogens_sel = hydrogens_sel
        h.acceptors_sel = acceptors_sel
        h.run(**client_HydrogenBondAnalysis)
        return h

    def test_hbond_modified_consistency(self, h, h_modified):
        assert_array_equal(h.results.hbonds, h_modified.results.hbonds)

    def test_hbond_analysis(self, h):

        assert len(np.unique(h.results.hbonds[:, 0])) == 10
        assert len(h.results.hbonds) == 32

        reference = {
            "distance": {"mean": 2.7627309, "std": 0.0905052},
            "angle": {"mean": 158.9038039, "std": 12.0362826},
        }

        assert_allclose(
            np.mean(h.results.hbonds[:, 4]), reference["distance"]["mean"]
        )
        assert_allclose(
            np.std(h.results.hbonds[:, 4]), reference["distance"]["std"]
        )
        assert_allclose(
            np.mean(h.results.hbonds[:, 5]), reference["angle"]["mean"]
        )
        assert_allclose(
            np.std(h.results.hbonds[:, 5]), reference["angle"]["std"]
        )

    def test_count_by_time(self, h):

        ref_times = np.arange(0.02, 0.21, 0.02)
        ref_counts = np.array([3, 2, 4, 4, 4, 4, 3, 2, 3, 3])

        counts = h.count_by_time()
        assert_array_almost_equal(h.times, ref_times)
        assert_array_equal(counts, ref_counts)

    def test_count_by_type(self, h):

        # Only one type of hydrogen bond in this system
        ref_count = 32

        counts = h.count_by_type()
        assert int(counts[0, 2]) == ref_count

    def test_count_by_ids(self, h, universe):

        ref_counts = [1.0, 1.0, 0.5, 0.4, 0.2, 0.1]
        unique_hbonds = h.count_by_ids()

        most_common_hbond_ids = [12, 14, 9]
        assert_equal(unique_hbonds[0, :3], most_common_hbond_ids)

        # count_by_ids() returns raw counts
        # convert to fraction of time that bond was observed
        counts = unique_hbonds[:, 3] / len(h.times)

        assert_allclose(counts, ref_counts)

    def test_hbonds_deprecated_attr(self, h):
        wmsg = "The `hbonds` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(h.hbonds, h.results.hbonds)


class TestHydrogenBondAnalysisIdeal(object):

    kwargs = {
        "donors_sel": "name O",
        "hydrogens_sel": "name H1 H2",
        "acceptors_sel": "name O",
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        # create two water molecules
        """
                       H4
                        \
            O1-H2 .... O2-H3
           /
          H1
        """
        n_residues = 2
        u = MDAnalysis.Universe.empty(
            n_atoms=n_residues * 3,
            n_residues=n_residues,
            atom_resindex=np.repeat(range(n_residues), 3),
            residue_segindex=[0] * n_residues,
            trajectory=True,  # necessary for adding coordinates
        )

        u.add_TopologyAttr("name", ["O", "H1", "H2"] * n_residues)
        u.add_TopologyAttr("type", ["O", "H", "H"] * n_residues)
        u.add_TopologyAttr("resname", ["SOL"] * n_residues)
        u.add_TopologyAttr("resid", list(range(1, n_residues + 1)))
        u.add_TopologyAttr("id", list(range(1, (n_residues * 3) + 1)))

        # Atomic coordinates with a single hydrogen bond between O1-H2---O2
        pos1 = np.array(
            [
                [0, 0, 0],  # O1
                [-0.249, -0.968, 0],  # H1
                [1, 0, 0],  # H2
                [2.5, 0, 0],  # O2
                [3.0, 0, 0],  # H3
                [2.250, 0.968, 0],  # H4
            ]
        )

        # Atomic coordinates with no hydrogen bonds
        pos2 = np.array(
            [
                [0, 0, 0],  # O1
                [-0.249, -0.968, 0],  # H1
                [1, 0, 0],  # H2
                [4.5, 0, 0],  # O2
                [5.0, 0, 0],  # H3
                [4.250, 0.968, 0],  # H4
            ]
        )

        coordinates = np.empty((3, u.atoms.n_atoms, 3))  # number of frames
        coordinates[0] = pos1
        coordinates[1] = pos2
        coordinates[2] = pos1
        u.load_new(coordinates, order="fac")

        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def hydrogen_bonds(universe, client_HydrogenBondAnalysis):
        h = HydrogenBondAnalysis(
            universe, **TestHydrogenBondAnalysisIdeal.kwargs
        )
        h.run(**client_HydrogenBondAnalysis)
        return h

    def test_count_by_type(self, hydrogen_bonds):

        # Only one type of hydrogen bond in this system
        ref_count = 2

        counts = hydrogen_bonds.count_by_type()
        assert int(counts[0, 2]) == ref_count

    def test_no_bond_info_exception(self, universe):

        kwargs = {
            "donors_sel": None,
            "hydrogens_sel": None,
            "acceptors_sel": None,
            "d_h_cutoff": 1.2,
            "d_a_cutoff": 3.0,
            "d_h_a_angle_cutoff": 120.0,
        }

        u = universe.copy()
        n_residues = 2
        u.add_TopologyAttr("mass", [15.999, 1.008, 1.008] * n_residues)
        u.add_TopologyAttr("charge", [-1.04, 0.52, 0.52] * n_residues)
        with pytest.raises(NoDataError, match="no bond information"):
            h = HydrogenBondAnalysis(u, **kwargs)

    def test_no_bond_donor_sel(self, universe):

        kwargs = {
            "donors_sel": "type O",
            "hydrogens_sel": None,
            "acceptors_sel": None,
            "d_h_cutoff": 1.2,
            "d_a_cutoff": 3.0,
            "d_h_a_angle_cutoff": 120.0,
        }
        u = universe.copy()
        n_residues = 2
        u.add_TopologyAttr("mass", [15.999, 1.008, 1.008] * n_residues)
        u.add_TopologyAttr("charge", [-1.04, 0.52, 0.52] * n_residues)
        h = HydrogenBondAnalysis(u, **kwargs)
        donors = u.select_atoms(h.guess_donors())

    def test_first_hbond(self, hydrogen_bonds):
        assert len(hydrogen_bonds.results.hbonds) == 2
        (
            frame_no,
            donor_index,
            hydrogen_index,
            acceptor_index,
            da_dst,
            angle,
        ) = hydrogen_bonds.results.hbonds[0]
        assert_equal(donor_index, 0)
        assert_equal(hydrogen_index, 2)
        assert_equal(acceptor_index, 3)
        assert_almost_equal(da_dst, 2.5)
        assert_almost_equal(angle, 180)

    def test_count_by_time(self, hydrogen_bonds):
        ref_times = np.array([0, 1, 2])  # u.trajectory.dt is 1
        ref_counts = np.array([1, 0, 1])

        counts = hydrogen_bonds.count_by_time()
        assert_array_almost_equal(hydrogen_bonds.times, ref_times)
        assert_array_equal(counts, ref_counts)

    def test_hydrogen_bond_lifetime(self, hydrogen_bonds):
        tau_timeseries, timeseries = hydrogen_bonds.lifetime(tau_max=2)
        assert_array_equal(timeseries, [1, 0, 0])

    def test_hydrogen_bond_lifetime_intermittency(self, hydrogen_bonds):
        tau_timeseries, timeseries = hydrogen_bonds.lifetime(
            tau_max=2, intermittency=1
        )
        assert_array_equal(timeseries, 1)

    def test_no_attr_hbonds(self, universe):
        hbonds = HydrogenBondAnalysis(universe, **self.kwargs)
        # hydrogen bonds are not computed
        with pytest.raises(NoDataError, match=".hbonds attribute is None"):
            hbonds.lifetime(tau_max=2, intermittency=1)

    def test_logging_step_not_1(
        self, universe, caplog, client_HydrogenBondAnalysis
    ):
        hbonds = HydrogenBondAnalysis(universe, **self.kwargs)
        # using step 2
        hbonds.run(**client_HydrogenBondAnalysis, step=2)

        caplog.set_level(logging.WARNING)
        hbonds.lifetime(tau_max=2, intermittency=1)

        warning = (
            "Autocorrelation: Hydrogen bonds were computed with " "step > 1."
        )
        assert any(warning in rec.getMessage() for rec in caplog.records)


class TestHydrogenBondAnalysisNoRes(TestHydrogenBondAnalysisIdeal):

    kwargs = {
        "donors_sel": "type O",
        #        'hydrogens_sel': 'type H H',
        "acceptors_sel": "type O",
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @staticmethod
    @pytest.fixture(scope="class", autouse=True)
    def universe():
        # create two water molecules
        """
                       H4
                        \
            O1-H2 .... O2-H3
           /
          H1
        """
        n_residues = 2
        u = MDAnalysis.Universe.empty(
            n_atoms=n_residues * 3,
            n_residues=n_residues,
            atom_resindex=np.repeat(range(n_residues), 3),
            residue_segindex=[0] * n_residues,
            trajectory=True,  # necessary for adding coordinates
        )

        u.add_TopologyAttr("type", ["O", "H", "H"] * n_residues)
        u.add_TopologyAttr("id", list(range(1, (n_residues * 3) + 1)))
        u.add_TopologyAttr("mass", [15.999, 1.008, 1.008] * n_residues)
        u.add_TopologyAttr("charge", [-1.04, 0.52, 0.52] * n_residues)

        # Atomic coordinates with a single hydrogen bond between O1-H2---O2
        pos1 = np.array(
            [
                [0, 0, 0],  # O1
                [-0.249, -0.968, 0],  # H1
                [1, 0, 0],  # H2
                [2.5, 0, 0],  # O2
                [3.0, 0, 0],  # H3
                [2.250, 0.968, 0],  # H4
            ]
        )

        # Atomic coordinates with no hydrogen bonds
        pos2 = np.array(
            [
                [0, 0, 0],  # O1
                [-0.249, -0.968, 0],  # H1
                [1, 0, 0],  # H2
                [4.5, 0, 0],  # O2
                [5.0, 0, 0],  # H3
                [4.250, 0.968, 0],  # H4
            ]
        )

        coordinates = np.empty((3, u.atoms.n_atoms, 3))  # number of frames
        coordinates[0] = pos1
        coordinates[1] = pos2
        coordinates[2] = pos1
        u.load_new(coordinates, order="fac")

        return u

    @staticmethod
    @pytest.fixture(scope="class")
    def hydrogen_bonds(universe, client_HydrogenBondAnalysis):
        h = HydrogenBondAnalysis(
            universe, **TestHydrogenBondAnalysisNoRes.kwargs
        )
        h.run(**client_HydrogenBondAnalysis)
        return h

    def test_no_hydrogen_bonds(self, universe):
        tmp_kwargs = copy.deepcopy(self.kwargs)
        tmp_kwargs["d_h_a_angle_cutoff"] = 50
        hbonds = HydrogenBondAnalysis(universe, **tmp_kwargs)

        with pytest.warns(
            UserWarning,
            match=(
                "No hydrogen bonds were found given angle "
                "of 50 between Donor, type O, and Acceptor,"
                " type O."
            ),
        ):
            hbonds.run(step=1)


class TestHydrogenBondAnalysisBetween(object):

    kwargs = {
        "donors_sel": "name O P",
        "hydrogens_sel": "name H1 H2 PH",
        "acceptors_sel": "name O P",
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        # create two water molecules and two "protein" molecules
        # P1-PH1 are the two atoms that comprise the toy protein PROT1
        """
                       H4
                        \
            O1-H2 .... O2-H3 ... P1-PH1 ... P2-PH2
           /
          H1
        """
        n_residues = 4
        n_sol_residues = 2
        n_prot_residues = 2
        u = MDAnalysis.Universe.empty(
            n_atoms=n_sol_residues * 3 + n_prot_residues * 2,
            n_residues=n_residues,
            atom_resindex=[0, 0, 0, 1, 1, 1, 2, 2, 3, 3],
            residue_segindex=[0, 0, 1, 1],
            trajectory=True,  # necessary for adding coordinates
        )

        u.add_TopologyAttr(
            "name",
            ["O", "H1", "H2"] * n_sol_residues + ["P", "PH"] * n_prot_residues,
        )
        u.add_TopologyAttr(
            "type",
            ["O", "H", "H"] * n_sol_residues + ["P", "PH"] * n_prot_residues,
        )
        u.add_TopologyAttr(
            "resname", ["SOL"] * n_sol_residues + ["PROT"] * n_prot_residues
        )
        u.add_TopologyAttr("resid", list(range(1, n_residues + 1)))
        u.add_TopologyAttr(
            "id",
            list(range(1, (n_sol_residues * 3 + n_prot_residues * 2) + 1)),
        )

        # Atomic coordinates with hydrogen bonds between:
        #     O1-H2---O2
        #     O2-H3---P1
        #     P1-PH1---P2
        pos = np.array(
            [
                [0, 0, 0],  # O1
                [-0.249, -0.968, 0],  # H1
                [1, 0, 0],  # H2
                [2.5, 0, 0],  # O2
                [3.0, 0, 0],  # H3
                [2.250, 0.968, 0],  # H4
                [5.5, 0, 0],  # P1
                [6.5, 0, 0],  # PH1
                [8.5, 0, 0],  # P2
                [9.5, 0, 0],  # PH2
            ]
        )

        coordinates = np.empty((1, u.atoms.n_atoms, 3))  # number of frames
        coordinates[0] = pos
        u.load_new(coordinates, order="fac")

        return u

    def test_between_all(self, universe, client_HydrogenBondAnalysis):
        # don't specify groups between which to find hydrogen bonds
        hbonds = HydrogenBondAnalysis(universe, between=None, **self.kwargs)
        hbonds.run(**client_HydrogenBondAnalysis)

        # indices of [donor, hydrogen, acceptor] for each hydrogen bond
        expected_hbond_indices = [
            [0, 2, 3],  # water-water
            [3, 4, 6],  # protein-water
            [6, 7, 8],  # protein-protein
        ]
        expected_hbond_distances = [2.5, 3.0, 3.0]
        assert_array_equal(
            hbonds.results.hbonds[:, 1:4], expected_hbond_indices
        )
        assert_allclose(hbonds.results.hbonds[:, 4], expected_hbond_distances)

    def test_between_PW(self, universe, client_HydrogenBondAnalysis):
        # Find only protein-water hydrogen bonds
        hbonds = HydrogenBondAnalysis(
            universe, between=["resname PROT", "resname SOL"], **self.kwargs
        )
        hbonds.run(**client_HydrogenBondAnalysis)

        # indices of [donor, hydrogen, acceptor] for each hydrogen bond
        expected_hbond_indices = [[3, 4, 6]]  # protein-water
        expected_hbond_distances = [3.0]
        assert_array_equal(
            hbonds.results.hbonds[:, 1:4], expected_hbond_indices
        )
        assert_allclose(hbonds.results.hbonds[:, 4], expected_hbond_distances)

    def test_between_PW_PP(self, universe, client_HydrogenBondAnalysis):
        # Find protein-water and protein-protein hydrogen bonds (not
        # water-water)
        hbonds = HydrogenBondAnalysis(
            universe,
            between=[
                ["resname PROT", "resname SOL"],
                ["resname PROT", "resname PROT"],
            ],
            **self.kwargs,
        )
        hbonds.run(**client_HydrogenBondAnalysis)

        # indices of [donor, hydrogen, acceptor] for each hydrogen bond
        expected_hbond_indices = [
            [3, 4, 6],  # protein-water
            [6, 7, 8],  # protein-protein
        ]
        expected_hbond_distances = [3.0, 3.0]
        assert_array_equal(
            hbonds.results.hbonds[:, 1:4], expected_hbond_indices
        )
        assert_allclose(hbonds.results.hbonds[:, 4], expected_hbond_distances)


class TestHydrogenBondAnalysisTIP3P_GuessAcceptors_GuessHydrogens_UseTopology_(
    TestHydrogenBondAnalysisTIP3P
):
    """Uses the same distance and cutoff hydrogen bond criteria as :class:`TestHydrogenBondAnalysisTIP3P`, so the
    results are identical, but the hydrogens and acceptors are guessed whilst the donor-hydrogen pairs are determined
    via the topology.
    """

    kwargs = {
        "donors_sel": None,
        "hydrogens_sel": None,
        "acceptors_sel": None,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    def test_no_hydrogens(self, universe, client_HydrogenBondAnalysis):
        # If no hydrogens are identified at a given frame, check an
        # empty donor atom group is created
        test_kwargs = TestHydrogenBondAnalysisTIP3P.kwargs.copy()
        test_kwargs["donors_sel"] = None  # use topology to find pairs
        test_kwargs["hydrogens_sel"] = "name H"  # no atoms have name H

        h = HydrogenBondAnalysis(universe, **test_kwargs)
        h.run(**client_HydrogenBondAnalysis)

        assert h._hydrogens.n_atoms == 0
        assert h._donors.n_atoms == 0
        assert h.results.hbonds.size == 0


class TestHydrogenBondAnalysisTIP3P_GuessDonors_NoTopology(object):
    """Guess the donor atoms involved in hydrogen bonds using the partial charges of the atoms."""

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        "donors_sel": None,
        "hydrogens_sel": None,
        "acceptors_sel": None,
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @pytest.fixture(scope="class")
    def h(self, universe):
        h = HydrogenBondAnalysis(universe, **self.kwargs)
        return h

    def test_guess_donors(self, h):

        ref_donors = "(resname TIP3 and name OH2)"
        donors = h.guess_donors(select="all", max_charge=-0.5)
        assert donors == ref_donors


class TestHydrogenBondAnalysisTIP3P_GuessHydrogens_NoTopology(object):
    """
    Guess the hydrogen atoms involved in hydrogen bonds using the mass and
    partial charge of the atoms.
    """

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        "donors_sel": None,
        "hydrogens_sel": None,
        "acceptors_sel": None,
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @pytest.fixture(scope="class")
    def h(self, universe):
        h = HydrogenBondAnalysis(universe, **self.kwargs)
        return h

    def test_guess_hydrogens(self, h):

        ref_hydrogens = (
            "(resname TIP3 and name H1) or (resname TIP3 and name H2)"
        )
        hydrogens = h.guess_hydrogens(select="all")
        assert hydrogens == ref_hydrogens

    pytest.mark.parametrize(
        "min_mass, max_mass, min_charge",
        [(1.05, 1.10, 0.30), (0.90, 0.95, 0.30), (0.90, 1.10, 1.00)],
    )

    def test_guess_hydrogens_empty_selection(self, h):

        hydrogens = h.guess_hydrogens(select="all", min_charge=1.0)
        assert hydrogens == ""

    def test_guess_hydrogens_min_max_mass(self, h):

        errmsg = r"min_mass is higher than \(or equal to\) max_mass"

        with pytest.raises(ValueError, match=errmsg):

            h.guess_hydrogens(select="all", min_mass=1.1, max_mass=0.9)


class TestHydrogenBondAnalysisTIP3PStartStep(object):
    """Uses the same distance and cutoff hydrogen bond criteria as :class:`TestHydrogenBondAnalysisTIP3P` but starting
    with the second frame and using every other frame in the analysis.
    """

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    kwargs = {
        "donors_sel": "name OH2",
        "hydrogens_sel": "name H1 H2",
        "acceptors_sel": "name OH2",
        "d_h_cutoff": 1.2,
        "d_a_cutoff": 3.0,
        "d_h_a_angle_cutoff": 120.0,
    }

    @pytest.fixture(scope="class")
    def h(self, universe, client_HydrogenBondAnalysis):
        h = HydrogenBondAnalysis(universe, **self.kwargs)
        h.run(**client_HydrogenBondAnalysis, start=1, step=2)
        return h

    def test_hbond_analysis(self, h):

        assert len(np.unique(h.results.hbonds[:, 0])) == 5
        assert len(h.results.hbonds) == 15

        reference = {
            "distance": {"mean": 2.73942464, "std": 0.05867924},
            "angle": {"mean": 157.07768079, "std": 9.72636682},
        }

        assert_allclose(
            np.mean(h.results.hbonds[:, 4]), reference["distance"]["mean"]
        )
        assert_allclose(
            np.std(h.results.hbonds[:, 4]), reference["distance"]["std"]
        )
        assert_allclose(
            np.mean(h.results.hbonds[:, 5]), reference["angle"]["mean"]
        )
        assert_allclose(
            np.std(h.results.hbonds[:, 5]), reference["angle"]["std"]
        )

    def test_count_by_time(self, h):

        ref_times = np.array(
            [
                0.04,
                0.08,
                0.12,
                0.16,
                0.20,
            ]
        )
        ref_counts = np.array([2, 4, 4, 2, 3])

        counts = h.count_by_time()
        assert_array_almost_equal(h.times, ref_times)
        assert_array_equal(counts, ref_counts)

    def test_count_by_type(self, h):

        # Only one type of hydrogen bond in this system
        ref_count = 15

        counts = h.count_by_type()
        assert int(counts[0, 2]) == ref_count


class TestHydrogenBondAnalysisEmptySelections:

    @staticmethod
    @pytest.fixture(scope="class")
    def universe():
        return MDAnalysis.Universe(waterPSF, waterDCD)

    msg = (
        "{} is an empty selection string - no hydrogen bonds will "
        "be found. This may be intended, but please check your "
        "selection."
    )

    @pytest.mark.parametrize(
        "seltype", ["donors_sel", "hydrogens_sel", "acceptors_sel"]
    )
    def test_empty_sel(self, universe, seltype):
        sel_kwarg = {seltype: " "}
        with pytest.warns(UserWarning, match=self.msg.format(seltype)):
            HydrogenBondAnalysis(universe, **sel_kwarg)

    def test_hbond_analysis(self, universe, client_HydrogenBondAnalysis):

        h = HydrogenBondAnalysis(
            universe, donors_sel=" ", hydrogens_sel=" ", acceptors_sel=" "
        )
        h.run(**client_HydrogenBondAnalysis)

        assert h.donors_sel == ""
        assert h.hydrogens_sel == ""
        assert h.acceptors_sel == ""
        assert h.results.hbonds.size == 0
